"""
Module that contains the state of transformation of TransformToGfm.
"""

from typing import List, Optional

from pymarkdown.tokens.markdown_token import MarkdownToken


# pylint: disable=too-many-instance-attributes
class TransformState:
    """
    Class that contains the state of transformation of TransformToGfm.
    """

    def __init__(self, actual_tokens: List[MarkdownToken]) -> None:
        """
        Initializes a new instance of the TransformState class.
        """
        (
            self.__is_in_code_block,
            self.__is_in_fenced_code_block,
            self.__is_in_html_block,
            self.__is_in_setext_block,
            self.__is_in_loose_list,
            self.__actual_tokens,
            self.__actual_token_index,
        ) = (False, False, False, False, True, actual_tokens, 0)
        self.__add_trailing_text: Optional[str] = None
        self.__add_leading_text: Optional[str] = None
        self.__next_token: Optional[MarkdownToken] = None
        self.__transform_stack: List[str] = []
        self.__last_token: Optional[MarkdownToken] = None

    @property
    def is_in_code_block(self) -> bool:
        """
        Whether the generator is currently inside of a code block.
        """
        return self.__is_in_code_block

    @is_in_code_block.setter
    def is_in_code_block(self, value: bool) -> None:
        """
        Set whether the generator is currently inside of a code block.
        """
        self.__is_in_code_block = value

    @property
    def is_in_setext_block(self) -> bool:
        """
        Whether the generator is currently inside of a setext block.
        """
        return self.__is_in_setext_block

    @is_in_setext_block.setter
    def is_in_setext_block(self, value: bool) -> None:
        """
        Set whether the generator is currently inside of a setext block.
        """
        self.__is_in_setext_block = value

    @property
    def is_in_fenced_code_block(self) -> bool:
        """
        Whether the generator is currently inside of a fenced code block.
        """
        return self.__is_in_fenced_code_block

    @is_in_fenced_code_block.setter
    def is_in_fenced_code_block(self, value: bool) -> None:
        """
        Set whether the generator is currently inside of a fenced code block.
        """
        self.__is_in_fenced_code_block = value

    @property
    def is_in_html_block(self) -> bool:
        """
        Whether the generator is currently inside of a HTML block.
        """
        return self.__is_in_html_block

    @is_in_html_block.setter
    def is_in_html_block(self, value: bool) -> None:
        """
        Set whether the generator is currently inside of a HTML block.
        """
        self.__is_in_html_block = value

    @property
    def is_in_loose_list(self) -> bool:
        """
        Whether the generator is currently inside of a loose list.
        """
        return self.__is_in_loose_list

    @is_in_loose_list.setter
    def is_in_loose_list(self, value: bool) -> None:
        """
        Set whether the generator is currently inside of a loose list.
        """
        self.__is_in_loose_list = value

    @property
    def transform_stack(self) -> List[str]:
        """
        Stack used to keep track of scope within the generator.
        """
        return self.__transform_stack

    @property
    def add_trailing_text(self) -> Optional[str]:
        """
        Keep track of trailing text.
        """
        return self.__add_trailing_text

    @add_trailing_text.setter
    def add_trailing_text(self, value: Optional[str]) -> None:
        """
        Set trailing text to keep track of.
        """
        self.__add_trailing_text = value

    @property
    def add_leading_text(self) -> Optional[str]:
        """
        Keep track of leading text.
        """
        return self.__add_leading_text

    @add_leading_text.setter
    def add_leading_text(self, value: Optional[str]) -> None:
        """
        Set leading text to keep track of.
        """
        self.__add_leading_text = value

    @property
    def next_token(self) -> Optional[MarkdownToken]:
        """
        Next token to process.
        """
        return self.__next_token

    @next_token.setter
    def next_token(self, value: Optional[MarkdownToken]) -> None:
        """
        Sets the next token to process.
        """
        self.__next_token = value

    @property
    def last_token(self) -> Optional[MarkdownToken]:
        """
        Last token to process.
        """
        return self.__last_token

    @last_token.setter
    def last_token(self, value: Optional[MarkdownToken]) -> None:
        """
        Sets the last token to process.
        """
        self.__last_token = value

    @property
    def actual_tokens(self) -> List[MarkdownToken]:
        """
        Actual tokens to use to generate the HTML with.
        """
        return self.__actual_tokens

    @property
    def actual_token_index(self) -> int:
        """
        Index into the actual token list.
        """
        return self.__actual_token_index

    @actual_token_index.setter
    def actual_token_index(self, value: int) -> None:
        """
        Sets the new index into the actual token list.
        """
        self.__actual_token_index = value


# pylint: enable=too-many-instance-attributes



"""
Module to implement a plugin that looks for excessive spaces after the block quote character.
"""

from typing import Dict, List, Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import PluginDetailsV2
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.plugins.utils.leading_space_index_tracker import (
    LeadingSpaceIndexTracker,
)
from pymarkdown.plugins.utils.list_tracker import ListTracker
from pymarkdown.tokens.blank_line_markdown_token import BlankLineMarkdownToken
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.fenced_code_block_markdown_token import (
    FencedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.inline_code_span_markdown_token import (
    InlineCodeSpanMarkdownToken,
)
from pymarkdown.tokens.link_reference_definition_markdown_token import (
    LinkReferenceDefinitionMarkdownToken,
)
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.new_list_item_markdown_token import NewListItemMarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.tokens.raw_html_markdown_token import RawHtmlMarkdownToken
from pymarkdown.tokens.setext_heading_markdown_token import SetextHeadingMarkdownToken
from pymarkdown.tokens.table_markdown_tokens import (
    TableMarkdownHeaderToken,
    TableMarkdownRowToken,
)
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


# pylint: disable=too-many-lines
# pylint: disable=too-many-instance-attributes
class RuleMd027(RulePlugin):
    """
    Class to implement a plugin that looks for excessive spaces after the block quote character.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__container_tokens: List[MarkdownToken] = []
        self.__bq_line_index: Dict[int, int] = {}
        self.__last_leaf_token: Optional[MarkdownToken] = None
        self.__line_index_at_bq_start: Optional[int] = None
        self.__is_paragraph_end_delayed = False
        self.__delayed_blank_line: Optional[MarkdownToken] = None
        self.__delayed_blank_line_bq_index: Optional[int] = None
        self.__delayed_blank_line_with_list_end = False
        self.__delayed_blank_line_container_token: Optional[MarkdownToken] = None
        self.__have_incremented_for_this_line = False
        self.__last_token: Optional[MarkdownToken] = None
        # self.__debug_on = False
        self.__list_tracker = ListTracker()
        self.__leading_space_index_tracker = LeadingSpaceIndexTracker()
        self.__previous_tokens: List[MarkdownToken] = []
        self.__table_header_hash: str = ""

        self.__delayed_bleading_fixes: Dict[
            MarkdownToken, List[Tuple[int, str, bool, BlankLineMarkdownToken]]
        ] = {}

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV2(
            plugin_name="no-multiple-space-blockquote",
            plugin_id="MD027",
            plugin_enabled_by_default=True,
            plugin_description="Multiple spaces after blockquote symbol",
            plugin_version="0.5.2",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md027.md",
            plugin_supports_fix=True,
            plugin_fix_level=5,
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__container_tokens = []
        self.__bq_line_index = {}
        self.__last_leaf_token = None
        self.__line_index_at_bq_start = None
        self.__is_paragraph_end_delayed = False
        self.__delayed_blank_line = None
        self.__delayed_blank_line_bq_index = None
        self.__delayed_blank_line_container_token = None
        self.__delayed_blank_line_with_list_end = False
        self.__have_incremented_for_this_line = False
        self.__last_token = None
        self.__list_tracker.starting_new_file()
        self.__leading_space_index_tracker.clear()
        self.__previous_tokens.clear()
        self.__table_header_hash = ""

    def __fix_issue(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        alternate_token: Optional[MarkdownToken],
    ) -> bool:
        keep_going = True
        if alternate_token:
            keep_going = self.__fix_issue_alternate_token(context, alternate_token)
        elif (
            self.__last_leaf_token
            and self.__last_leaf_token.is_setext_heading
            and token.is_text
        ):
            self.__fix_issue_setext_text(context, token)
        elif (
            token.is_setext_heading
            or token.is_thematic_break
            or token.is_fenced_code_block
            or token.is_atx_heading
            or token.is_blank_line
        ):
            self.register_fix_token_request(
                context, token, "next_token", "extracted_whitespace", ""
            )
        elif token.is_new_list_item:
            self.__fix_issue_new_list_item(context, token)
        elif token.is_list_start:
            self.__fix_issue_list_start(context, token)
        elif token.is_table_header:
            self.__fix_issue_table_header(context, token)
        elif token.is_table_row:
            self.__fix_issue_table_row(context, token)
        else:
            assert token.is_link_reference_definition, "Unexpected token fix request."
            self.__fix_issue_link_reference(context, token)
        return keep_going

    # pylint: disable=too-many-arguments
    def __report_issue(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        line_number_delta: int = 0,
        column_number_delta: int = 0,
        alternate_token: Optional[MarkdownToken] = None,
    ) -> bool:
        if context.in_fix_mode:
            return self.__fix_issue(context, token, alternate_token)

        if token.is_setext_heading:
            current_line_number = cast(
                SetextHeadingMarkdownToken, token
            ).original_line_number
        else:
            current_line_number = token.line_number

        token_index = len(self.__previous_tokens) - 1
        if self.__previous_tokens[token_index].is_table:
            token_index -= 1
        while token_index >= 0 and self.__previous_tokens[token_index].is_container:
            token_index -= 1
        override_is_error_token_prefaced_by_blank_line = bool(
            (
                token_index >= 0
                and self.__previous_tokens[token_index].line_number
                == (current_line_number - 1)
                and self.__previous_tokens[token_index].is_blank_line
            )
        )
        self.report_next_token_error(
            context,
            token,
            line_number_delta=line_number_delta
            + context.calc_pragma_offset(token, line_number_delta),
            column_number_delta=column_number_delta,
            override_is_error_token_prefaced_by_blank_line=override_is_error_token_prefaced_by_blank_line,
        )
        return True

    # pylint: enable=too-many-arguments

    def __fix_issue_alternate_token(
        self,
        context: PluginScanContext,
        alternate_token: MarkdownToken,
    ) -> bool:
        if alternate_token.is_paragraph:
            para_token = cast(ParagraphMarkdownToken, alternate_token)
            extracted_whitespace = "\n" * ParserHelper.count_newlines_in_text(
                para_token.extracted_whitespace
            )
            self.register_fix_token_request(
                context,
                alternate_token,
                "next_token",
                "extracted_whitespace",
                extracted_whitespace,
            )
            return False
        assert (
            alternate_token.is_setext_heading_end
            or alternate_token.is_fenced_code_block_end
        )
        self.register_fix_token_request(
            context, alternate_token, "next_token", "extracted_whitespace", ""
        )
        return True

    def __fix_issue_list_start(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        list_start_token = cast(ListStartMarkdownToken, token)
        adjust_amount = len(list_start_token.extracted_whitespace)
        self.register_fix_token_request(
            context, token, "next_token", "extracted_whitespace", ""
        )
        self.register_fix_token_request(
            context,
            token,
            "next_token",
            "indent_level",
            list_start_token.indent_level - adjust_amount,
        )
        self.register_fix_token_request(
            context,
            token,
            "next_token",
            "column_number",
            list_start_token.column_number - adjust_amount,
        )
        self.__list_tracker.register(token, adjust_amount)

    def __fix_issue_link_reference(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        assert token.is_link_reference_definition
        lrd_token = cast(LinkReferenceDefinitionMarkdownToken, token)
        if lrd_token.extracted_whitespace:
            self.register_fix_token_request(
                context, token, "next_token", "extracted_whitespace", ""
            )
        assert lrd_token.link_destination_whitespace

        modified_whitespace = "\n" * ParserHelper.count_newlines_in_text(
            lrd_token.link_destination_whitespace
        )
        if modified_whitespace != lrd_token.link_destination_whitespace:
            self.register_fix_token_request(
                context,
                token,
                "next_token",
                "link_destination_whitespace",
                modified_whitespace,
            )

        if lrd_token.link_title_whitespace:
            modified_whitespace = "\n" * ParserHelper.count_newlines_in_text(
                lrd_token.link_title_whitespace
            )
            if modified_whitespace != lrd_token.link_title_whitespace:
                self.register_fix_token_request(
                    context,
                    token,
                    "next_token",
                    "link_title_whitespace",
                    modified_whitespace,
                )

    def __process_delayed_paragraph_end(
        self, token: MarkdownToken, num_container_tokens: int
    ) -> None:
        if self.__is_paragraph_end_delayed:
            # if self.__debug_on:
            #     print("[[Processing delayed paragraph end]]")
            #     print(f"delay-para-end-->token->{ParserHelper.make_value_visible(token)}")
            #     print(f"delay-para-end-->index->{self.__bq_line_index[num_container_tokens]}")

            ff = False
            assert (
                token.is_blank_line
                or token.is_thematic_break
                or token.is_atx_heading
                # setext
                # icb
                or token.is_fenced_code_block
                or token.is_html_block
                # lrd
                # para
                # tables
                or token.is_list_start
                or token.is_block_quote_end
                or ff
            )
            self.__bq_line_index[num_container_tokens] += 1
            self.__have_incremented_for_this_line = True
            # if self.__debug_on:
            #     print(f"delay-para-end-->index->{self.__bq_line_index[num_container_tokens]}")
            #     print("[[Delayed paragraph end processed]]")
            self.__is_paragraph_end_delayed = False

    # pylint: disable=too-many-arguments
    def __process_delayed_blank_line(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
        allow_block_quote_end: bool,
        is_block_quote_end: bool,
    ) -> None:
        if (
            self.__delayed_blank_line
            and not (
                token.is_leaf_end_token
                or (allow_block_quote_end and token.is_block_quote_end)
            )
            and (
                not self.__have_incremented_for_this_line
                or token.is_blank_line
                or is_block_quote_end
            )
        ):
            self.__have_incremented_for_this_line = False
            assert self.__delayed_blank_line_bq_index is not None
            self.__handle_blank_line(
                context,
                self.__delayed_blank_line,
                num_container_tokens,
                self.__delayed_blank_line_bq_index,
            )
            self.__delayed_blank_line = None
            self.__delayed_blank_line_bq_index = None
            self.__delayed_blank_line_with_list_end = False
            self.__delayed_blank_line_container_token = None

    # pylint: enable=too-many-arguments

    def __process_pending_container_end_tokens(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        _ = context
        did_reduce_containers = False
        while self.__leading_space_index_tracker.have_any_registered_container_ends():
            self.__leading_space_index_tracker.process_container_end(token)
            did_reduce_containers = True

        if (
            did_reduce_containers
            and not self.__leading_space_index_tracker.in_at_least_one_container()
        ):
            self.__previous_tokens.clear()

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        # if self.__debug_on:
        #     print(f">>{ParserHelper.make_value_visible(token)}")
        if (
            self.__have_incremented_for_this_line
            and not token.is_end_token
            and not token.is_blank_line
        ):
            self.__have_incremented_for_this_line = False

        if not token.is_end_token or token.is_end_of_stream:
            self.__process_pending_container_end_tokens(context, token)

        num_container_tokens = len(
            [i for i in self.__container_tokens if i.is_block_quote_start]
        )
        # if self.__debug_on:
        #     print(f"num->bqs:{num_container_tokens}, self.__have_incremented_for_this_line=" + \
        #       f"{self.__have_incremented_for_this_line}")
        #     if num_container_tokens in self.__bq_line_index:
        #         print(f"{self.__bq_line_index[num_container_tokens]}-->token>" + \
        #           f"{ParserHelper.make_value_visible(token)}")
        #     else:
        #         print(f"token>{ParserHelper.make_value_visible(token)}")
        #     if self.__container_tokens:
        #         print(f"self.__container_tokens>{ParserHelper.make_value_visible(self.__container_tokens)}")
        if token.is_block_quote_start:
            self.__leading_space_index_tracker.open_container(token)
            self.__handle_block_quote_start(token)
        elif token.is_block_quote_end:
            self.__leading_space_index_tracker.register_container_end(token)
            self.__handle_block_quote_end(context, token, num_container_tokens)
        elif token.is_list_start:
            self.__leading_space_index_tracker.open_container(token)
            self.__handle_list_start(context, token, num_container_tokens)
        elif token.is_list_end:
            self.__leading_space_index_tracker.register_container_end(token)
            self.__handle_list_end(context, token)
        elif token.is_new_list_item:
            self.__handle_new_list_item(context, token, num_container_tokens)
        else:
            self.__list_tracker.next_token(token)
            if num_container_tokens:
                self.__handle_within_block_quotes(context, token)

        self.__leading_space_index_tracker.track_since_last_non_end_token(token)

        self.__last_token = token
        self.__previous_tokens.append(token)

    def __handle_block_quote_start(self, token: MarkdownToken) -> None:
        self.__container_tokens.append(token)

        num_container_tokens = len(
            [i for i in self.__container_tokens if i.is_block_quote_start]
        )
        self.__bq_line_index[num_container_tokens] = 0
        self.__is_paragraph_end_delayed = False
        # if self.__debug_on:
        #     print(f"bq>{ParserHelper.make_value_visible(self.__container_tokens[-1])}")

    # pylint: disable=too-many-arguments
    def __register_blank_line(
        self,
        token: BlockQuoteMarkdownToken,
        index: int,
        mod: str,
        did_x: bool,
        blank_line_token: BlankLineMarkdownToken,
    ) -> None:
        if token in self.__delayed_bleading_fixes:
            delayed_list = self.__delayed_bleading_fixes[token]
        else:
            delayed_list = []
            self.__delayed_bleading_fixes[token] = delayed_list
        delayed_list.append((index, mod, did_x, blank_line_token))

    # pylint: enable=too-many-arguments

    def __handle_block_quote_end(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
    ) -> None:
        self.__process_delayed_paragraph_end(token, num_container_tokens)
        self.__process_delayed_blank_line(
            context, token, num_container_tokens, False, True
        )

        num_container_tokens = len(
            [i for i in self.__container_tokens if i.is_block_quote_start]
        )
        # if self.__debug_on:
        #     print(f"leading_spaces>{ParserHelper.make_value_visible(self.__container_tokens[-1].leading_spaces)}")
        block_quote_token = cast(BlockQuoteMarkdownToken, self.__container_tokens[-1])
        assert block_quote_token.bleading_spaces is not None

        if (
            self.__delayed_bleading_fixes
            and block_quote_token in self.__delayed_bleading_fixes
        ):
            split_leading_spaces = block_quote_token.bleading_spaces.split(
                ParserHelper.newline_character
            )
            delayed_list = self.__delayed_bleading_fixes[block_quote_token]
            for delayed_list_item in delayed_list:
                block_quote_index = delayed_list_item[0]
                modified_part = delayed_list_item[1]
                split_leading_spaces[block_quote_index] = modified_part

                did_trigger = delayed_list_item[2]
                if not did_trigger:
                    blank_token = delayed_list_item[3]
                    self.report_next_token_error(
                        context,
                        blank_token,
                        column_number_delta=-(blank_token.column_number - 1),
                    )
            if context.in_fix_mode:
                self.register_fix_token_request(
                    context,
                    block_quote_token,
                    "next_token",
                    "bleading_spaces",
                    "\n".join(split_leading_spaces),
                )
            del self.__delayed_bleading_fixes[block_quote_token]

        newlines_in_container = block_quote_token.bleading_spaces.count(
            ParserHelper.newline_character
        )
        if (
            not (
                block_quote_token.bleading_spaces
                and block_quote_token.bleading_spaces.endswith(
                    ParserHelper.newline_character
                )
            )
            and self.__container_tokens
            and block_quote_token.bleading_spaces
        ):
            # if self.__debug_on:
            #     print(f"newlines_in_container>{newlines_in_container}")
            newlines_in_container += 1

        # if self.__debug_on:
        #     print(f"newlines_in_container>{newlines_in_container}")
        #     print(f"__bq_line_index>{self.__bq_line_index[num_container_tokens]}")
        #     print(f"__is_paragraph_end_delayed>{self.__is_paragraph_end_delayed}")
        #     print(f"__delayed_blank_line>{self.__delayed_blank_line}")
        # if (
        #     self.__delayed_blank_line
        #     and newlines_in_container != self.__bq_line_index[num_container_tokens]
        # ):
        #     self.__bq_line_index[num_container_tokens] += 1

        # assert newlines_in_container == self.__bq_line_index[num_container_tokens], (
        #     str(newlines_in_container)
        #     + " == "
        #     + str(self.__bq_line_index[num_container_tokens])
        # )
        del self.__bq_line_index[num_container_tokens]
        del self.__container_tokens[-1]
        end_token = cast(EndMarkdownToken, token)
        if end_token.extra_end_data and num_container_tokens > 1:
            self.__bq_line_index[num_container_tokens - 1] -= 1

    def __get_last_block_quote(self) -> Optional[MarkdownToken]:
        return next(
            (
                self.__container_tokens[i]
                for i in range(len(self.__container_tokens) - 1, -1, -1)
                if self.__container_tokens[i].is_block_quote_start
            ),
            None,
        )

    def __check_list_starts(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
        is_new_list_item: bool,
    ) -> None:
        # if self.__debug_on:
        #     print(f"num_container_tokens={num_container_tokens};")
        #     print(f"container_tokens={ParserHelper.make_value_visible(self.__container_tokens)};")
        _ = num_container_tokens
        if found_block_quote_token := self.__get_last_block_quote():
            # is_start_properly_scoped = False
            is_start_properly_scoped = (
                (found_block_quote_token == self.__container_tokens[-2])
                if is_new_list_item
                else (found_block_quote_token == self.__container_tokens[-1])
            )
            # if self.__debug_on:
            #     print(
            #         f"is_start_properly_scoped={is_start_properly_scoped};"
            #         + f"found_block_quote_token={ParserHelper.make_value_visible(found_block_quote_token)}"
            #     )
            #     print(f"token.extracted_whitespace>:{token.extracted_whitespace}:")
            #     print(f"self.__last_token>:{ParserHelper.make_value_visible(self.__last_token)}:")

            # In rare cases, a block quote is ended in the middle of a line.
            # This is one of those cases.
            list_token = cast(ListStartMarkdownToken, token)
            whitespace_to_use = list_token.extracted_whitespace
            if (
                self.__last_token
                and self.__last_token.is_end_token
                and self.__last_token.is_block_quote_end
            ):
                end_token = cast(EndMarkdownToken, self.__last_token)
                block_quote_token = cast(
                    BlockQuoteMarkdownToken, end_token.start_markdown_token
                )
                # if self.__debug_on:
                #     print(f"self.__last_token.start_markdown_token>:{ParserHelper.make_value_visible(\
                #       self.__last_token.start_markdown_token)}:")
                #     print("BOOM")
                assert block_quote_token.bleading_spaces is not None
                split_line_length = block_quote_token.bleading_spaces.split("\n")[-1]
                # if self.__debug_on:
                #     print(f"BOOM:{split_line_length}:")
                whitespace_to_use = whitespace_to_use[len(split_line_length) :]

            if is_start_properly_scoped and whitespace_to_use:
                column_number_delta = -(token.column_number - len(whitespace_to_use))
                # if self.__debug_on:
                #     print("list-error")
                #     line_index = (
                #         self.__bq_line_index[num_container_tokens] + 0
                #     )
                #     print(f"5>{line_index}")
                #     print(f"column-delta>{column_number_delta}")
                self.__report_issue(
                    context, token, column_number_delta=column_number_delta
                )

    def __handle_list_start(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
    ) -> None:
        self.__list_tracker.list_start(token)

        self.__process_delayed_blank_line(
            context, token, num_container_tokens, False, False
        )
        self.__process_delayed_paragraph_end(token, num_container_tokens)
        self.__check_list_starts(context, token, num_container_tokens, False)
        self.__container_tokens.append(token)

    def __fix_issue_new_list_item(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        list_start_token = cast(NewListItemMarkdownToken, token)
        adjust_amount = len(list_start_token.extracted_whitespace)
        self.register_fix_token_request(
            context, token, "next_token", "extracted_whitespace", ""
        )
        self.register_fix_token_request(
            context,
            token,
            "next_token",
            "indent_level",
            list_start_token.indent_level - adjust_amount,
        )
        # self.register_fix_token_request(context, token, "next_token", "column_number", list_start_token.column_number - adjust_amount)
        self.__list_tracker.register(token, adjust_amount)

    def __handle_new_list_item(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
    ) -> None:
        self.__list_tracker.new_list_item(token)
        # if self.__debug_on:
        #     print(
        #         f"num_container_tokens={num_container_tokens}, __is_paragraph_end_delayed="
        #         + f"{self.__is_paragraph_end_delayed}, self.__have_incremented_for_this_line="
        #         + f"{self.__have_incremented_for_this_line}"
        #     )
        if (
            num_container_tokens
            and not self.__have_incremented_for_this_line
            and self.__is_paragraph_end_delayed
        ):
            self.__bq_line_index[num_container_tokens] += 1
            self.__is_paragraph_end_delayed = False
        self.__check_list_starts(context, token, num_container_tokens, True)

    def __handle_list_end(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        self.__list_tracker.list_end()
        if self.__delayed_blank_line:
            self.__delayed_blank_line_with_list_end = True

        if registration_map := self.__list_tracker.get_registrations():
            end_token = cast(EndMarkdownToken, token)
            list_token = cast(ListStartMarkdownToken, end_token.start_markdown_token)
            if list_token.leading_spaces:
                split_leading_spaces = list_token.leading_spaces.split("\n")
                for registered_token, adj in registration_map.items():
                    start, stop = self.__list_tracker.get_start_stop(registered_token)
                    for next_index in range(start, stop):
                        assert adj > 0
                        split_leading_spaces[next_index] = split_leading_spaces[
                            next_index
                        ][:-adj]
                        # else:
                        #     split_leading_spaces[next_index] = split_leading_spaces[next_index] + (" " * -adj)
                rebuilt_leading_spaces = "\n".join(split_leading_spaces)
                if rebuilt_leading_spaces != list_token.leading_spaces:
                    self.register_fix_token_request(
                        context,
                        list_token,
                        "next_token",
                        "leading_spaces",
                        rebuilt_leading_spaces,
                    )

        self.__list_tracker.list_end_cleanup()
        del self.__container_tokens[-1]

    def __handle_blank_line(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
        delayed_bq_index: int,
    ) -> None:
        # if self.__debug_on:
        #     print(f"__handle_blank_line>>{token}<<")
        blank_line_token = cast(BlankLineMarkdownToken, token)

        assert self.__delayed_blank_line_container_token is not None
        scoped_token = self.__delayed_blank_line_container_token
        if scoped_token is not None and scoped_token.is_block_quote_start:

            self.__handle_blank_line_inner(
                scoped_token, delayed_bq_index, blank_line_token, token
            )

        if blank_line_token.extracted_whitespace:
            # if self.__debug_on:
            #     print("blank-error")

            self.__report_issue(context, token)

        assert self.__bq_line_index
        self.__bq_line_index[num_container_tokens] += 1

    def __handle_blank_line_inner(
        self,
        scoped_token: MarkdownToken,
        delayed_bq_index: int,
        blank_line_token: BlankLineMarkdownToken,
        token: MarkdownToken,
    ) -> None:
        scoped_block_quote_token = cast(BlockQuoteMarkdownToken, scoped_token)
        assert scoped_block_quote_token.bleading_spaces is not None
        split_leading_spaces = scoped_block_quote_token.bleading_spaces.split(
            ParserHelper.newline_character
        )

        container_index = (
            self.__leading_space_index_tracker.get_container_stack_size() - 1
        )
        if self.__leading_space_index_tracker.get_container_stack_item(
            container_index
        ).is_block_quote_start:
            block_quote_index = self.__leading_space_index_tracker.get_tokens_block_quote_bleading_space_index(
                token
            )
        else:
            block_quote_index = delayed_bq_index

        # If we are closing other containers, can cause issues. So do not fire.
        is_end_with_other_closed_containers = (
            block_quote_index == len(split_leading_spaces) - 1
            and self.__delayed_blank_line_with_list_end
        )

        # If we have matching nested block quotes, and then a blank line, we need to prevent
        # the firing.
        is_special_case = (
            block_quote_index == 0
            and len(split_leading_spaces) == 1
            and blank_line_token.line_number != scoped_block_quote_token.line_number
        )

        assert (
            not is_special_case
            and not is_end_with_other_closed_containers
            and block_quote_index < len(split_leading_spaces)
        )

        specific_block_quote_prefix = split_leading_spaces[block_quote_index]
        mod_specific_block_quote_prefix = specific_block_quote_prefix.rstrip(" ")
        if mod_specific_block_quote_prefix != specific_block_quote_prefix:
            self.__register_blank_line(
                scoped_block_quote_token,
                block_quote_index,
                mod_specific_block_quote_prefix,
                bool(blank_line_token.extracted_whitespace),
                blank_line_token,
            )

    def __handle_common_element(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
        is_directly_within_block_quote: bool,
    ) -> None:
        common_token = cast(BlankLineMarkdownToken, token)
        if common_token.extracted_whitespace and is_directly_within_block_quote:
            column_number_delta = -(
                token.column_number - len(common_token.extracted_whitespace)
            )
            self.__report_issue(context, token, column_number_delta=column_number_delta)
        self.__bq_line_index[num_container_tokens] += 1
        self.__last_leaf_token = token

    def __handle_atx_heading_or_thematic_break(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
        is_directly_within_block_quote: bool,
    ) -> None:
        self.__handle_common_element(
            context, token, num_container_tokens, is_directly_within_block_quote
        )

    def __fix_issue_setext_text(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        text_token = cast(TextMarkdownToken, token)
        whitespace_parts: List[str] = []
        assert text_token.end_whitespace is not None
        for next_line in text_token.end_whitespace.split("\n"):
            split_character_index = next_line.find(
                ParserHelper.whitespace_split_character
            )
            if split_character_index == -1:
                whitespace_parts.append(next_line)
            else:
                whitespace_parts.append(next_line[split_character_index + 1 :])
        recombined_whitespace = "\n".join(whitespace_parts)
        assert recombined_whitespace != text_token.end_whitespace
        self.register_fix_token_request(
            context, token, "next_token", "end_whitespace", recombined_whitespace
        )

    def __handle_setext_heading(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        is_directly_within_block_quote: bool,
    ) -> None:
        setext_token = cast(SetextHeadingMarkdownToken, token)
        if setext_token.extracted_whitespace and is_directly_within_block_quote:
            line_number_delta = setext_token.original_line_number - token.line_number
            column_number_delta = -(
                setext_token.original_column_number
                - len(setext_token.extracted_whitespace)
            )
            # if self.__debug_on:
            #     print("setext-error")
            #     print(f"line->{token.original_line_number},col={token.original_column_number}")
            #     print(f"delta->line->{line_number_delta},col={column_number_delta}")
            self.__report_issue(
                context,
                token,
                line_number_delta=line_number_delta,
                column_number_delta=column_number_delta,
            )
        self.__last_leaf_token = token

    def __handle_setext_heading_end(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
        is_directly_within_block_quote: bool,
    ) -> None:
        end_token = cast(EndMarkdownToken, token)
        if end_token.extracted_whitespace and is_directly_within_block_quote:
            # if self.__debug_on:
            #     print("setext_heading_end-error")
            assert self.__last_leaf_token is not None
            column_number_delta = -(
                self.__last_leaf_token.column_number
                - len(end_token.extracted_whitespace)
            )
            assert self.__last_leaf_token is not None
            self.__report_issue(
                context,
                self.__last_leaf_token,
                column_number_delta=column_number_delta,
                alternate_token=token,
            )
        self.__bq_line_index[num_container_tokens] += 1
        self.__last_leaf_token = None

    def __handle_fenced_code_block(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
        is_directly_within_block_quote: bool,
    ) -> None:
        self.__handle_common_element(
            context, token, num_container_tokens, is_directly_within_block_quote
        )
        self.__line_index_at_bq_start = self.__bq_line_index[num_container_tokens]

    def __handle_fenced_code_block_end(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
        is_directly_within_block_quote: bool,
    ) -> None:  # sourcery skip: extract-method
        fenced_token = cast(FencedCodeBlockMarkdownToken, token)
        if fenced_token.extracted_whitespace and is_directly_within_block_quote:
            scoped_block_quote_token = cast(
                BlockQuoteMarkdownToken, self.__container_tokens[-1]
            )
            assert scoped_block_quote_token.bleading_spaces is not None
            split_leading_spaces = scoped_block_quote_token.bleading_spaces.split(
                ParserHelper.newline_character
            )
            specific_block_quote_prefix = split_leading_spaces[
                self.__bq_line_index[num_container_tokens]
            ]

            assert self.__line_index_at_bq_start is not None
            line_number_delta = 1 + (
                self.__bq_line_index[num_container_tokens]
                - self.__line_index_at_bq_start
            )
            column_number_delta = -(len(specific_block_quote_prefix) + 1)

            # if self.__debug_on:
            #     print(f"end-container>>{ParserHelper.make_value_visible(self.__container_tokens[-1])}")
            #     print(f"split_leading_spaces>>{ParserHelper.make_value_visible(split_leading_spaces)}")
            #     print("specific_block_quote_prefix>>:" + \
            #       f"{ParserHelper.make_value_visible(specific_block_quote_prefix)}:")
            #     print("fenced-end-error")
            assert self.__last_leaf_token is not None
            self.__report_issue(
                context,
                self.__last_leaf_token,
                line_number_delta=line_number_delta,
                column_number_delta=column_number_delta,
                alternate_token=token,
            )
        self.__last_leaf_token = None
        self.__bq_line_index[num_container_tokens] += 1

    # pylint: disable=too-many-arguments
    def __report_lrd_error(
        self,
        lrd_token: LinkReferenceDefinitionMarkdownToken,
        num_container_tokens: int,
        context: PluginScanContext,
        token: MarkdownToken,
        scoped_block_quote_token: BlockQuoteMarkdownToken,
    ) -> None:
        assert lrd_token.link_name_debug is not None
        assert scoped_block_quote_token.bleading_spaces is not None
        line_number_delta = (
            lrd_token.link_name_debug.count(ParserHelper.newline_character) + 1
        )

        split_array_index = (
            self.__bq_line_index[num_container_tokens] + line_number_delta
        )
        split_leading_spaces = scoped_block_quote_token.bleading_spaces.split(
            ParserHelper.newline_character
        )
        specific_block_quote_prefix = split_leading_spaces[split_array_index]

        column_number_delta = -(len(specific_block_quote_prefix) + 1)

        # if self.__debug_on:
        #     print(f"line_number_delta>>{line_number_delta}")
        #     print(f"split_array_index>>{split_array_index}")
        #     print(f"end-container>>{ParserHelper.make_value_visible(self.__container_tokens[-1])}")
        #     print(f"split_leading_spaces>>{ParserHelper.make_value_visible(split_leading_spaces)}")
        #     print("specific_block_quote_prefix>>:" + \
        #       f"{ParserHelper.make_value_visible(specific_block_quote_prefix)}:")
        #     print("lrd-2-error")
        self.__report_issue(
            context,
            token,
            line_number_delta=line_number_delta,
            column_number_delta=column_number_delta,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def __handle_link_reference_definition_check_after(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
        lrd_token: LinkReferenceDefinitionMarkdownToken,
        leading_spaces: str,
    ) -> None:
        assert lrd_token.link_name_debug is not None
        assert lrd_token.link_title_whitespace is not None
        line_number_delta = (
            lrd_token.link_name_debug.count(ParserHelper.newline_character)
            + lrd_token.link_title_whitespace.count(ParserHelper.newline_character)
            + 1
        )

        split_array_index = (
            self.__bq_line_index[num_container_tokens] + line_number_delta
        )
        split_leading_spaces = leading_spaces.split(ParserHelper.newline_character)
        specific_block_quote_prefix = split_leading_spaces[split_array_index]

        column_number_delta = -(len(specific_block_quote_prefix) + 1)
        # if self.__debug_on:
        #     print("line_number_delta>>" + str(line_number_delta))
        #     print("split_array_index>>" + str(split_array_index))
        #     print(f"end-container>>{ParserHelper.make_value_visible(self.__container_tokens[-1])}")
        #     print(f"split_leading_spaces>>{ParserHelper.make_value_visible(split_leading_spaces)}")
        #     print("specific_block_quote_prefix>>:" + \
        #       f"{ParserHelper.make_value_visible(specific_block_quote_prefix)}:")
        #     print("lrd-3-error")
        self.__report_issue(
            context,
            token,
            line_number_delta=line_number_delta,
            column_number_delta=column_number_delta,
        )

    # pylint: enable=too-many-arguments

    def __handle_code_span(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        code_span_token = cast(InlineCodeSpanMarkdownToken, token)

        start_index = 0
        next_index = code_span_token.span_text.find("\x07\n\x07", start_index)
        recombine_list: List[str] = []
        while next_index != -1:
            recombine_list.extend(
                (code_span_token.span_text[start_index:next_index], "\n")
            )
            after_space_index, _ = ParserHelper.collect_while_spaces_verified(
                code_span_token.span_text, next_index + 3
            )
            is_there = ParserHelper.is_character_at_index(
                code_span_token.span_text, after_space_index, "\x07"
            )
            assert is_there
            start_index = after_space_index + 1
            next_index = code_span_token.span_text.find("\x07\n\x07", start_index)
        recombine_list.append(code_span_token.span_text[start_index:])
        recombined_span_text = "".join(recombine_list)
        if recombined_span_text != code_span_token.span_text:
            self.register_fix_token_request(
                context, token, "next_token", "span_text", recombined_span_text
            )

    def __handle_raw_html(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        raw_html_token = cast(RawHtmlMarkdownToken, token)
        recombine_list: List[str] = []
        for next_tag_line in raw_html_token.raw_tag.split("\n"):
            if next_tag_line.startswith("\a"):
                after_space_index, _ = ParserHelper.collect_while_spaces_verified(
                    next_tag_line, 1
                )
                remaining_line = next_tag_line[after_space_index:]
                assert remaining_line.startswith("\a\x03\a")
                remaining_line = remaining_line[3:]
                recombine_list.append(remaining_line)
            else:
                recombine_list.append(next_tag_line)
        recombined_raw_tag = "\n".join(recombine_list)
        if recombined_raw_tag != raw_html_token.raw_tag:
            self.register_fix_token_request(
                context, token, "next_token", "raw_tag", recombined_raw_tag
            )

    def __handle_link_reference_definition(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
    ) -> None:
        scoped_block_quote_token = cast(
            BlockQuoteMarkdownToken, self.__container_tokens[-1]
        )
        if scoped_block_quote_token.is_block_quote_start:
            assert scoped_block_quote_token.bleading_spaces is not None
            leading_spaces = scoped_block_quote_token.bleading_spaces
        else:
            scoped_list_token = cast(
                ListStartMarkdownToken, self.__container_tokens[-1]
            )
            assert scoped_list_token.leading_spaces is not None
            leading_spaces = scoped_list_token.leading_spaces
        lrd_token = cast(LinkReferenceDefinitionMarkdownToken, token)
        if lrd_token.extracted_whitespace:
            column_number_delta = -(
                lrd_token.column_number - len(lrd_token.extracted_whitespace)
            )
            # if self.__debug_on:
            #     print("lrd-1-error")
            self.__report_issue(context, token, column_number_delta=column_number_delta)

        assert lrd_token.link_destination_whitespace is not None
        found_index = lrd_token.link_destination_whitespace.find(
            ParserHelper.newline_character
        )
        if found_index != -1 and ParserHelper.is_character_at_index_whitespace(
            lrd_token.link_destination_whitespace, found_index + 1
        ):
            self.__report_lrd_error(
                lrd_token,
                num_container_tokens,
                context,
                token,
                scoped_block_quote_token,
            )

        assert lrd_token.link_title_whitespace is not None
        found_index = lrd_token.link_title_whitespace.find(
            ParserHelper.newline_character
        )
        if found_index != -1 and ParserHelper.is_character_at_index_whitespace(
            lrd_token.link_title_whitespace, found_index + 1
        ):
            self.__handle_link_reference_definition_check_after(
                context, token, num_container_tokens, lrd_token, leading_spaces
            )

        assert lrd_token.link_name_debug is not None
        assert lrd_token.link_title_raw is not None
        self.__bq_line_index[num_container_tokens] += (
            1
            + lrd_token.link_name_debug.count(ParserHelper.newline_character)
            + lrd_token.link_destination_whitespace.count(
                ParserHelper.newline_character
            )
            + lrd_token.link_title_whitespace.count(ParserHelper.newline_character)
            + lrd_token.link_title_raw.count(ParserHelper.newline_character)
        )

    # pylint: disable=too-many-arguments
    def __scan_text(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        next_line: str,
        scoped_block_quote_token: BlockQuoteMarkdownToken,
        num_container_tokens: int,
        line_number_delta: int,
    ) -> None:
        found_index = next_line.find(ParserHelper.whitespace_split_character)
        if found_index != -1:
            next_line = next_line[:found_index]
        if next_line and found_index != -1:
            assert scoped_block_quote_token.bleading_spaces is not None
            split_leading_spaces = scoped_block_quote_token.bleading_spaces.split(
                ParserHelper.newline_character
            )
            split_array_index = (
                self.__bq_line_index[num_container_tokens] + line_number_delta + 1
            )
            specific_block_quote_prefix = split_leading_spaces[split_array_index]
            calculated_column_number = -(len(specific_block_quote_prefix) + 1)

            # if self.__debug_on:
            #     print(f"split_leading_spaces>>{ParserHelper.make_value_visible(split_leading_spaces)}")
            #     print(f"split_array_index>>{ParserHelper.make_value_visible(split_array_index)}")
            #     print("specific_block_quote_prefix>>:" + \
            #       f"{ParserHelper.make_value_visible(specific_block_quote_prefix)}:")
            #     print("setext-text-error")
            self.__report_issue(
                context,
                token,
                line_number_delta=line_number_delta,
                column_number_delta=calculated_column_number,
            )

    # pylint: enable=too-many-arguments

    def __handle_text(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
        is_directly_within_block_quote: bool,
    ) -> None:
        assert self.__last_leaf_token is not None
        text_token = cast(TextMarkdownToken, token)
        if self.__last_leaf_token.is_setext_heading:
            if is_directly_within_block_quote:
                scoped_block_quote_token = cast(
                    BlockQuoteMarkdownToken, self.__container_tokens[-1]
                )
                assert scoped_block_quote_token.bleading_spaces is not None
                assert text_token.end_whitespace is not None

                for line_number_delta, next_line in enumerate(
                    text_token.end_whitespace.split(ParserHelper.newline_character)
                ):
                    self.__scan_text(
                        context,
                        token,
                        next_line,
                        scoped_block_quote_token,
                        num_container_tokens,
                        line_number_delta,
                    )
            assert text_token.end_whitespace is not None
            self.__bq_line_index[num_container_tokens] += (
                text_token.end_whitespace.count(ParserHelper.newline_character) + 1
            )
        elif (
            self.__last_leaf_token.is_html_block or self.__last_leaf_token.is_code_block
        ):
            self.__bq_line_index[num_container_tokens] += (
                text_token.token_text.count(ParserHelper.newline_character) + 1
            )

    def __handle_paragraph(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
        is_directly_within_block_quote: bool,
    ) -> None:
        paragraph_token = cast(ParagraphMarkdownToken, token)
        self.__last_leaf_token = paragraph_token
        if is_directly_within_block_quote:
            scoped_block_quote_token = cast(
                BlockQuoteMarkdownToken, self.__container_tokens[-1]
            )
            # if self.__debug_on:
            #     print(f"para>>>{scoped_block_quote_token}")
            delta_x = paragraph_token.line_number - scoped_block_quote_token.line_number

            for paragraph_start_line_number_delta, next_line in enumerate(
                paragraph_token.extracted_whitespace.split(
                    ParserHelper.newline_character
                )
            ):
                if next_line and scoped_block_quote_token.bleading_spaces:
                    # if self.__debug_on:
                    #     print(f"1>{self.__bq_line_index[num_container_tokens]}")
                    #     print(f"2>{line_number_delta}")
                    #     print(f"3>{ParserHelper.make_value_visible(scoped_block_quote_token)}")
                    split_leading_spaces = (
                        scoped_block_quote_token.bleading_spaces.split(
                            ParserHelper.newline_character
                        )
                    )
                    line_index = (
                        self.__bq_line_index[num_container_tokens]
                        + paragraph_start_line_number_delta
                    )
                    calculated_column_number = len(split_leading_spaces[line_index]) + 1
                    # if self.__debug_on:
                    #     print(f"4>{split_leading_spaces}")
                    #     print(f"5>{line_index}")
                    #     print(f"column>{calculated_column_number}")
                    #     print("para-error")
                    keep_going = self.__report_issue(
                        context,
                        scoped_block_quote_token,
                        line_number_delta=paragraph_start_line_number_delta + delta_x,
                        column_number_delta=-calculated_column_number,
                        alternate_token=paragraph_token,
                    )
                    if not keep_going:
                        break
        self.__bq_line_index[
            num_container_tokens
        ] += paragraph_token.extracted_whitespace.count(ParserHelper.newline_character)

    def __handle_within_block_quotes_prefix(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> Tuple[int, bool]:
        # if self.__debug_on:
        #     print("__handle_within_block_quotes")
        num_container_tokens = len(
            [i for i in self.__container_tokens if i.is_block_quote_start]
        )
        # if self.__debug_on:
        #     print(f"{self.__bq_line_index}-->index>{num_container_tokens}")
        #     print(f"{self.__bq_line_index[num_container_tokens]}-->token>{ParserHelper.make_value_visible(token)}")

        self.__process_delayed_paragraph_end(token, num_container_tokens)
        self.__process_delayed_blank_line(
            context, token, num_container_tokens, True, False
        )
        is_directly_within_block_quote = self.__container_tokens[
            -1
        ].is_block_quote_start
        # if self.__debug_on:
        #     print(f"{self.__bq_line_index[num_container_tokens]}-->token>{ParserHelper.make_value_visible(token)}")
        return num_container_tokens, is_directly_within_block_quote

    def __handle_within_block_quotes(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        (
            num_container_tokens,
            is_directly_within_block_quote,
        ) = self.__handle_within_block_quotes_prefix(context, token)

        if token.is_paragraph:
            self.__handle_paragraph(
                context, token, num_container_tokens, is_directly_within_block_quote
            )
        elif token.is_text:
            self.__handle_text(
                context, token, num_container_tokens, is_directly_within_block_quote
            )
        elif token.is_paragraph_end:
            self.__last_leaf_token = None
            self.__is_paragraph_end_delayed = True
            # if self.__debug_on:
            #     print("[[Delaying paragraph end]]")
        elif token.is_blank_line:
            self.__delayed_blank_line = token
            self.__delayed_blank_line_bq_index = self.__bq_line_index[
                num_container_tokens
            ]
            self.__delayed_blank_line_with_list_end = False
            self.__delayed_blank_line_container_token = self.__container_tokens[-1]
            # if self.__debug_on:
            #     print("[[Delaying blank line]]")
        elif token.is_atx_heading or token.is_thematic_break:
            self.__handle_atx_heading_or_thematic_break(
                context, token, num_container_tokens, is_directly_within_block_quote
            )
        elif token.is_setext_heading:
            self.__handle_setext_heading(context, token, is_directly_within_block_quote)
        elif token.is_setext_heading_end:
            self.__handle_setext_heading_end(
                context, token, num_container_tokens, is_directly_within_block_quote
            )
        elif token.is_atx_heading_end:
            self.__last_leaf_token = None
        elif token.is_link_reference_definition:
            self.__handle_link_reference_definition(
                context, token, num_container_tokens
            )
        else:
            self.__handle_within_block_quotes_remaining(
                context, token, num_container_tokens, is_directly_within_block_quote
            )
        # if self.__debug_on:
        #     print(f"{self.__bq_line_index[num_container_tokens]}<--token>{ParserHelper.make_value_visible(token)}")

    def __handle_within_block_quotes_remaining(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        num_container_tokens: int,
        is_directly_within_block_quote: bool,
    ) -> None:
        # add table support: https://github.com/jackdewinter/pymarkdown/issues/1515

        if token.is_table:
            self.__last_leaf_token = token
        elif token.is_table_header:
            self.__handle_table_header(context, token, is_directly_within_block_quote)
        elif token.is_table_row:
            self.__handle_table_row(context, token)
        elif token.is_table_end:
            self.__last_leaf_token = None
        elif token.is_inline_raw_html and context.in_fix_mode:
            self.__handle_raw_html(context, token)
        elif token.is_inline_code_span and context.in_fix_mode:
            self.__handle_code_span(context, token)
        else:
            self.__handle_within_block_quotes_blocks(
                token, context, num_container_tokens, is_directly_within_block_quote
            )

    def __fix_issue_table_header(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        assert token.is_table_header
        table_header_token = cast(TableMarkdownHeaderToken, token)

        # This is one of the few cases where we fix 2 things on one token, so it is more
        # convenient to check for this here instead of in the main code.  As there can be
        # two issues with the single header token and we cannot distinguish between the
        # two of them at fix time, we do it here by checking to see if we have already
        # fixed this token.
        new_hash = (
            f"{table_header_token.line_number}:{table_header_token.column_number}"
        )
        if new_hash == self.__table_header_hash:
            return

        if table_header_token.header_row_leading_whitespace:
            self.register_fix_token_request(
                context, token, "next_token", "header_row_leading_whitespace", ""
            )
        found_whitespace_characters, _ = ParserHelper.extract_spaces(
            table_header_token.separator_line, 0
        )
        if found_whitespace_characters:
            self.register_fix_token_request(
                context,
                token,
                "next_token",
                "separator_line",
                table_header_token.separator_line[found_whitespace_characters:],
            )
        self.__table_header_hash = new_hash

    def __handle_table_header(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        is_directly_within_block_quote: bool,
    ) -> None:
        assert is_directly_within_block_quote

        table_header_token = cast(TableMarkdownHeaderToken, token)
        if table_header_token.header_row_leading_whitespace:
            column_number_delta = -(
                table_header_token.column_number
                - len(table_header_token.header_row_leading_whitespace)
            )
            self.__report_issue(
                context,
                token,
                column_number_delta=column_number_delta,
            )
        found_whitespace_characters, _ = ParserHelper.extract_spaces(
            table_header_token.separator_line, 0
        )
        if found_whitespace_characters:
            column_number_delta = -(
                table_header_token.column_number
                - len(table_header_token.header_row_leading_whitespace)
            )
            assert self.__last_leaf_token and self.__last_leaf_token.is_table
            self.__report_issue(
                context,
                token,
                line_number_delta=1,
                column_number_delta=column_number_delta,
            )

    def __fix_issue_table_row(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        assert token.is_table_row
        table_row_token = cast(TableMarkdownRowToken, token)
        assert table_row_token.leading_whitespace
        self.register_fix_token_request(
            context, token, "next_token", "leading_whitespace", ""
        )

    def __handle_table_row(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
    ) -> None:
        table_row_token = cast(TableMarkdownRowToken, token)
        if table_row_token.leading_whitespace:
            column_number_delta = -(
                table_row_token.column_number - len(table_row_token.leading_whitespace)
            )
            self.__report_issue(
                context,
                token,
                column_number_delta=column_number_delta,
            )

    def __handle_within_block_quotes_blocks(
        self,
        token: MarkdownToken,
        context: PluginScanContext,
        num_container_tokens: int,
        is_directly_within_block_quote: bool,
    ) -> None:
        if token.is_fenced_code_block:
            self.__handle_fenced_code_block(
                context, token, num_container_tokens, is_directly_within_block_quote
            )
        elif token.is_fenced_code_block_end:
            self.__handle_fenced_code_block_end(
                context, token, num_container_tokens, is_directly_within_block_quote
            )
        elif token.is_html_block or token.is_indented_code_block:
            self.__last_leaf_token = token
        elif token.is_html_block_end or token.is_indented_code_block_end:
            self.__last_leaf_token = None


# pylint: enable=too-many-instance-attributes



"""
Module to provide an indication of an error while tokenizing the Markdown file.
"""


class BadTokenizationError(Exception):
    """
    Class to allow for a critical error during the tokenization of the Markdown file
    to be reported.
    """

    def __init__(self, formatted_message: str):
        super().__init__(formatted_message)



"""
Module to implement a plugin that looks for excessively long lines in the file.
"""

from typing import List, Optional, Tuple

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken


# pylint: disable=too-many-instance-attributes
class RuleMd013(RulePlugin):
    """
    Class to implement a plugin that looks for excessively long lines in the file.
    """

    __maximum_line_length = 99999

    def __init__(self) -> None:
        super().__init__()
        self.__leaf_tokens: List[MarkdownToken] = []
        self.__line_index = 0
        self.__leaf_token_index = 0
        self.__line_length = 0
        self.__code_block_line_length = 0
        self.__heading_line_length = 0
        self.__minimum_line_length = 0
        self.__code_blocks_active = False
        self.__headings_active = False
        self.__strict_mode = False
        self.__stern_mode = False
        self.__last_line: Optional[str] = None

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="line-length",
            plugin_id="MD013",
            plugin_enabled_by_default=True,
            plugin_description="Line length",
            plugin_version="0.6.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md013.md",
            plugin_configuration="line_length,heading_line_length,code_block_line_length,"
            + "code_blocks,headings,strict,stern",
        )

    @classmethod
    def __validate_minimum(cls, found_value: int) -> None:
        if found_value < 1:
            raise ValueError("Allowable values are any integer greater than 0.")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__line_length = (
            self.plugin_configuration.get_integer_property_with_default(
                "line_length",
                80,
                valid_value_fn=self.__validate_minimum,
            )
        )
        self.__code_block_line_length = (
            self.plugin_configuration.get_integer_property_with_default(
                "code_block_line_length",
                80,
                valid_value_fn=self.__validate_minimum,
            )
        )
        self.__heading_line_length = (
            self.plugin_configuration.get_integer_property_with_default(
                "heading_line_length",
                80,
                valid_value_fn=self.__validate_minimum,
            )
        )
        self.__minimum_line_length = min(
            self.__line_length,
            self.__code_block_line_length,
            self.__heading_line_length,
        )

        self.__code_blocks_active = (
            self.plugin_configuration.get_boolean_property_with_default(
                "code_blocks",
                True,
            )
        )
        self.__headings_active = (
            self.plugin_configuration.get_boolean_property_with_default(
                "headings",
                True,
            )
        )
        self.__strict_mode = (
            self.plugin_configuration.get_boolean_property_with_default(
                "strict",
                False,
            )
        )
        self.__stern_mode = self.plugin_configuration.get_boolean_property_with_default(
            "stern",
            False,
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("line_length", self.__line_length),
            QueryConfigItem("code_block_line_length", self.__code_block_line_length),
            QueryConfigItem("heading_line_length", self.__heading_line_length),
            QueryConfigItem("code_blocks", self.__code_blocks_active),
            QueryConfigItem("headings", self.__headings_active),
            QueryConfigItem("strict", self.__strict_mode),
            QueryConfigItem("stern", self.__stern_mode),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__leaf_tokens = []
        self.__line_index = 1
        self.__leaf_token_index = 0
        self.__last_line = None

    def __is_really_longer(
        self, line_length: int, compare_length: int
    ) -> Tuple[bool, int]:
        # print("line(" + str(self.__line_index) + ")->len=(" + str(line_length) + "):" + str(line))
        # print("-->" + str(self.__leaf_tokens[self.__leaf_token_index]))
        if (
            self.__leaf_tokens[self.__leaf_token_index].is_fenced_code_block
            or self.__leaf_tokens[self.__leaf_token_index].is_indented_code_block
        ):
            compare_length = (
                self.__code_block_line_length
                if self.__code_blocks_active
                else RuleMd013.__maximum_line_length
            )
        elif (
            self.__leaf_tokens[self.__leaf_token_index].is_atx_heading
            or self.__leaf_tokens[self.__leaf_token_index].is_setext_heading
        ):
            compare_length = (
                self.__heading_line_length
                if self.__headings_active
                else RuleMd013.__maximum_line_length
            )
        return line_length > compare_length, compare_length

    def next_line(self, context: PluginScanContext, line: str) -> None:
        """
        Event that a new line is being processed.
        """
        if (
            self.__leaf_token_index + 1 < len(self.__leaf_tokens)
            and self.__line_index
            == self.__leaf_tokens[self.__leaf_token_index + 1].line_number
        ):
            self.__leaf_token_index += 1

        line_length = len(line)
        compare_length = self.__line_length
        if line_length > self.__minimum_line_length:
            is_actually_longer, compare_length = self.__is_really_longer(
                line_length, compare_length
            )
        else:
            is_actually_longer = False
        if is_actually_longer:
            # trigger_rule = False
            if self.__strict_mode:
                trigger_rule = True
            else:
                next_space_index, _ = ParserHelper.extract_until_spaces(
                    line, compare_length
                )
                # print("next_index=" + str(next_space_index))
                trigger_rule = (
                    line_length == next_space_index
                    if self.__stern_mode
                    else line_length != next_space_index
                )

            if trigger_rule:
                extra_error_information = (
                    f"Expected: {compare_length}, Actual: {line_length}"
                )
                override_is_error_token_prefaced_by_blank_line = (
                    self.__last_line is not None and not self.__last_line
                )
                self.report_next_line_error(
                    context,
                    1,
                    extra_error_information=extra_error_information,
                    override_is_error_token_prefaced_by_blank_line=override_is_error_token_prefaced_by_blank_line,
                )
        self.__line_index += 1
        self.__last_line = line.strip(" \t")

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        _ = context
        if token.is_blank_line or token.is_leaf:
            self.__leaf_tokens.append(token)


# pylint: enable=too-many-instance-attributes



"""
Module to allow for the details on the extension to be encapsulated.
"""

from dataclasses import dataclass
from typing import Optional

# pylint: disable=too-many-instance-attributes


@dataclass(frozen=True)
class ExtensionDetails:
    """
    Class to allow for the details on the extension to be encapsulated.
    """

    extension_id: str
    extension_name: str
    extension_description: str
    extension_enabled_by_default: bool
    extension_version: str
    extension_interface_version: int
    extension_url: Optional[str]
    extension_configuration: Optional[str]


# pylint: enable=too-many-instance-attributes



"""
Module to provide helpers for the processing of leaf blocks.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import ListStackToken, StackToken

POGGER = ParserLogger(logging.getLogger(__name__))


class LeafBlockHelper:
    """
    Class to provide helpers for the processing of leaf blocks.
    """

    # pylint: disable=too-many-arguments
    @staticmethod
    def correct_for_leaf_block_start_in_list(
        parser_state: ParserState,
        removed_chars_at_start: int,
        old_top_of_stack_token: StackToken,
        html_tokens: List[MarkdownToken],
        block_quote_data: BlockQuoteData,
        was_token_already_added_to_stack: bool = True,
        delay_tab_match: bool = False,
        alt_removed_chars_at_start: Optional[int] = None,
        is_html: bool = False,
        original_line: Optional[str] = None,
    ) -> None:
        """
        Check to see that if a paragraph has been closed within a list and
        there is a leaf block token immediately following, that the right
        actions are taken.
        """

        POGGER.debug(
            ">>correct_for_leaf_block_start_in_list>>removed_chars_at_start>$>>",
            removed_chars_at_start,
        )
        if not old_top_of_stack_token.is_paragraph:
            POGGER.debug("1")
            return

        statck_index, top_of_stack, end_of_list = (
            -2 if was_token_already_added_to_stack else -1,
            None,
            html_tokens[-1],
        )
        if not parser_state.token_stack[statck_index].is_list:
            POGGER.debug("2")
            return

        POGGER.debug(
            ">>correct_for_leaf_block_start_in_list>>stack>>$>>",
            parser_state.token_stack,
        )
        POGGER.debug(
            ">>correct_for_leaf_block_start_in_list>>tokens>>$>>",
            parser_state.token_document,
        )
        POGGER.debug(
            ">>correct_for_leaf_block_start_in_list>>tokens_to_add>>$>>", html_tokens
        )

        if was_token_already_added_to_stack:
            top_of_stack = parser_state.token_stack[-1]
            del parser_state.token_stack[-1]
        del html_tokens[-1]

        LeafBlockHelper.__handle_leaf_start(
            parser_state,
            removed_chars_at_start,
            html_tokens,
            delay_tab_match,
            alt_removed_chars_at_start,
            is_html,
            block_quote_data,
            original_line,
        )

        if was_token_already_added_to_stack:
            assert (
                top_of_stack is not None
            ), "If token was added, top_of_stack should be defined."
            parser_state.token_stack.append(top_of_stack)
            POGGER.debug(
                ">>correct_for_leaf_block_start_in_list>>stack>>$>>",
                parser_state.token_stack,
            )
        html_tokens.append(end_of_list)
        POGGER.debug(
            ">>correct_for_leaf_block_start_in_list>>tokens_to_add>>$>>", html_tokens
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_leaf_start_adjust_tab(
        parser_state: ParserState,
        original_line: str,
        leading_chars_at_start: str,
        indent_count: int,
        used_indent: str,
    ) -> str:
        after_index_index = len(leading_chars_at_start) + indent_count
        used_indent_text = leading_chars_at_start + used_indent
        assert (
            parser_state.original_line_to_parse is not None
        ), "Original line must be defined by now."
        reconstructed_line = (
            used_indent_text + parser_state.original_line_to_parse[after_index_index:]
        )
        detabified_original_line = TabHelper.detabify_string(original_line)
        assert (
            detabified_original_line == reconstructed_line
        ), "Detabified original must equal reconstructed."
        keep_going = True
        detab_index = 1
        while keep_going:
            orig_prefix = original_line[:detab_index]
            detab_orig_prefix = TabHelper.detabify_string(orig_prefix)
            keep_going = len(detab_orig_prefix) < len(used_indent_text)
            if keep_going:
                detab_index += 1
        assert len(detab_orig_prefix) == len(
            used_indent_text
        ), "Detab prefix must equal used indent."
        return orig_prefix[len(leading_chars_at_start) :]

    @staticmethod
    def __determine_removed_and_leading_characters(
        parser_state: ParserState, removed_chars_at_start: int
    ) -> Tuple[int, str]:
        leading_chars_at_start = ""
        last_bq_index = parser_state.find_last_block_quote_on_stack()
        if last_bq_index > 0:
            last_bq_token = cast(
                BlockQuoteMarkdownToken,
                parser_state.token_stack[last_bq_index].matching_markdown_token,
            )
            assert (
                last_bq_token.bleading_spaces is not None
            ), "Block quote tokens must have bleading spaces defined"
            leading_chars_at_start = last_bq_token.bleading_spaces
            last_newline_index = leading_chars_at_start.rfind("\n")
            assert (
                last_newline_index != -1
            ), "leading_chars_at_start must always have at least one newline character"
            leading_chars_at_start = leading_chars_at_start[last_newline_index + 1 :]
            removed_chars_at_start -= len(leading_chars_at_start)
        return removed_chars_at_start, leading_chars_at_start

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_leaf_start_adjust(
        parser_state: ParserState,
        delay_tab_match: bool,
        alt_removed_chars_at_start: Optional[int],
        is_html: bool,
        original_line: Optional[str],
        list_stack_token: ListStackToken,
        removed_chars_at_start: int,
    ) -> None:
        if LeafBlockHelper.__detect_list_already_added_to(parser_state):
            POGGER.debug("__handle_leaf_start_adjust->__detect_list_already_added_to")
            return

        if delay_tab_match:
            POGGER.debug("__handle_leaf_start_adjust->delay_tab_match")
            used_indent = ""
        else:
            POGGER.debug("__handle_leaf_start_adjust->tab_match")
            leading_chars_at_start = ""
            if alt_removed_chars_at_start is not None:
                indent_count = alt_removed_chars_at_start
            else:
                if is_html:
                    removed_chars_at_start, leading_chars_at_start = (
                        LeafBlockHelper.__determine_removed_and_leading_characters(
                            parser_state, removed_chars_at_start
                        )
                    )
                indent_count = removed_chars_at_start
            used_indent = ParserHelper.repeat_string(" ", indent_count)
            if (
                is_html
                and original_line is not None
                and leading_chars_at_start
                and "\t" in original_line
            ):
                used_indent = LeafBlockHelper.__handle_leaf_start_adjust_tab(
                    parser_state,
                    original_line,
                    leading_chars_at_start,
                    indent_count,
                    used_indent,
                )
        assert (
            list_stack_token.matching_markdown_token is not None
        ), "Container stack tokens always have markdown tokens."
        list_markdown_token = cast(
            ListStartMarkdownToken, list_stack_token.matching_markdown_token
        )
        POGGER.debug("__handle_leaf_start_adjust>>list_token>>$", list_markdown_token)
        list_markdown_token.add_leading_spaces(used_indent)
        POGGER.debug("__handle_leaf_start_adjust>>list_token>>$", list_markdown_token)

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-locals
    @staticmethod
    def __detect_list_already_added_to(parser_state: ParserState) -> bool:

        assert len(parser_state.block_copy) == len(parser_state.copy_of_token_stack) - 1
        copy_stack_index = len(parser_state.copy_of_token_stack) - 1
        while (
            copy_stack_index > 0
            and not parser_state.copy_of_token_stack[copy_stack_index].is_list
            and not parser_state.copy_of_token_stack[copy_stack_index].is_block_quote
        ):
            copy_stack_index -= 1
        stack_index = len(parser_state.token_stack) - 1
        assert (
            parser_state.token_stack[stack_index].is_list
            or parser_state.token_stack[stack_index].is_block_quote
        )
        # while (
        #     stack_index > 0
        #     and not parser_state.token_stack[stack_index].is_list
        #     and not parser_state.token_stack[stack_index].is_block_quote
        # ):
        #     stack_index -= 1
        token_stack_markdown_token = parser_state.token_stack[
            stack_index
        ].matching_markdown_token
        assert token_stack_markdown_token is not None
        bb_line = token_stack_markdown_token.line_number
        bb_column = token_stack_markdown_token.column_number
        assert (
            parser_state.copy_of_token_stack[copy_stack_index].matching_markdown_token
            is not None
        )

        copy_of_token_stack_markdown_token = parser_state.copy_of_token_stack[
            copy_stack_index
        ].matching_markdown_token
        assert copy_of_token_stack_markdown_token is not None
        aa_line = copy_of_token_stack_markdown_token.line_number
        aa_column = copy_of_token_stack_markdown_token.column_number
        new_stack_index = copy_stack_index
        while new_stack_index > 0 and bb_line != aa_line and bb_column != aa_column:
            new_stack_index -= 1
            copy_of_token_stack_markdown_token = parser_state.copy_of_token_stack[
                new_stack_index
            ].matching_markdown_token
            assert copy_of_token_stack_markdown_token is not None
            aa_line = copy_of_token_stack_markdown_token.line_number
            aa_column = copy_of_token_stack_markdown_token.column_number

        original_removed_tokens: List[ListStartMarkdownToken] = []
        removed_tokens: List[ListStartMarkdownToken] = []
        add_index = new_stack_index + 1
        while add_index <= copy_stack_index:
            original_token = parser_state.copy_of_token_stack[
                add_index
            ].matching_markdown_token
            assert original_token is not None and original_token.is_list_start
            original_removed_tokens.append(cast(ListStartMarkdownToken, original_token))

            current_token = parser_state.block_copy[add_index - 1]
            assert current_token is not None and current_token.is_list_start
            removed_tokens.append(cast(ListStartMarkdownToken, current_token))
            add_index += 1

        assert original_removed_tokens
        assert len(original_removed_tokens) == 1
        assert original_removed_tokens[0].is_list_start

        original_leading_spaces = original_removed_tokens[0].leading_spaces
        current_leading_spaces = removed_tokens[0].leading_spaces
        return bool(
            (
                original_leading_spaces
                and current_leading_spaces
                and original_leading_spaces != current_leading_spaces
            )
        )

    # pylint: enable=too-many-locals

    @staticmethod
    def __handle_leaf_start_inner(
        parser_state: ParserState,
        removed_chars_at_start: int,
        ws_count: int,
        html_tokens: List[MarkdownToken],
        stack_delta: int,
    ) -> Tuple[Optional[ListStackToken], int]:
        assert parser_state.token_stack[
            -1
        ].is_list, "Token at the end of the stack must be a list token."
        list_stack_token = cast(ListStackToken, parser_state.token_stack[-1])

        POGGER.debug(">>removed_chars_at_start>>$>>", removed_chars_at_start)
        POGGER.debug(">>stack indent>>$>>", list_stack_token.indent_level)
        if (removed_chars_at_start + ws_count) >= list_stack_token.indent_level:
            return None, stack_delta
        tokens_from_close, _ = parser_state.close_open_blocks_fn(
            parser_state,
            until_this_index=(len(parser_state.token_stack) - 1),
            include_lists=True,
        )
        POGGER.debug(
            ">>correct_for_leaf_block_start_in_list>>tokens_from_close>>$>>",
            tokens_from_close,
        )
        html_tokens.extend(tokens_from_close)
        remaining_stack_index = len(parser_state.token_stack) - 1
        while (
            stack_delta > 0
            and parser_state.token_stack[remaining_stack_index].is_block_quote
        ):
            stack_delta -= 1
            remaining_stack_index -= 1
        if remaining_stack_index != len(parser_state.token_stack) - 1:
            tokens_from_close, _ = parser_state.close_open_blocks_fn(
                parser_state,
                until_this_index=remaining_stack_index + 1,
                include_block_quotes=True,
            )
            html_tokens.extend(tokens_from_close)
        return list_stack_token, stack_delta

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_leaf_start(
        parser_state: ParserState,
        removed_chars_at_start: int,
        html_tokens: List[MarkdownToken],
        delay_tab_match: bool,
        alt_removed_chars_at_start: Optional[int],
        is_html: bool,
        block_quote_data: BlockQuoteData,
        original_line: Optional[str] = None,
    ) -> None:
        POGGER.debug(
            ">>correct_for_leaf_block_start_in_list>>stack>>$>>",
            parser_state.token_stack,
        )
        POGGER.debug(
            ">>correct_for_leaf_block_start_in_list>>tokens_to_add>>$>>", html_tokens
        )

        assert parser_state.original_line_to_parse is not None
        ws_count, _ = ParserHelper.collect_while_spaces_verified(
            parser_state.original_line_to_parse[removed_chars_at_start:], 0
        )

        stack_delta = block_quote_data.stack_count - block_quote_data.current_count

        adjust_with_leading_spaces = False
        is_remaining_list_token = True
        while is_remaining_list_token:
            list_stack_token, stack_delta = LeafBlockHelper.__handle_leaf_start_inner(
                parser_state, removed_chars_at_start, ws_count, html_tokens, stack_delta
            )
            if list_stack_token is None:
                break
            adjust_with_leading_spaces = True
            is_remaining_list_token = parser_state.token_stack[-1].is_list

        POGGER.debug("is_remaining_list_token=$", is_remaining_list_token)
        if is_remaining_list_token:
            assert parser_state.token_stack[
                -1
            ].is_list, "Token at the end of the stack must be a list token."
            list_stack_token = cast(ListStackToken, parser_state.token_stack[-1])
            # delta_indent = removed_chars_at_start - list_stack_token.indent_level
            # POGGER.debug(
            #     ">>correct_for_leaf_block_start_in_list>>delta_indent>>$>>",
            #     delta_indent,
            # )
            # assert not delta_indent

            POGGER.debug(">>delay_tab_match>>$>>", delay_tab_match)
            POGGER.debug(
                ">>adjust_with_leading_spaces>>$>>", adjust_with_leading_spaces
            )
            if adjust_with_leading_spaces:
                LeafBlockHelper.__handle_leaf_start_adjust(
                    parser_state,
                    delay_tab_match,
                    alt_removed_chars_at_start,
                    is_html,
                    original_line,
                    list_stack_token,
                    removed_chars_at_start,
                )

    # pylint: enable=too-many-arguments

    @staticmethod
    def extract_markdown_tokens_back_to_blank_line(
        parser_state: ParserState, was_forced: bool
    ) -> List[MarkdownToken]:
        """
        Extract tokens going back to the last blank line token.
        """

        pre_tokens: List[MarkdownToken] = []
        while parser_state.token_document[-1].is_blank_line:
            last_element = parser_state.token_document[-1]
            if was_forced:
                pre_tokens.insert(0, last_element)
            else:
                pre_tokens.append(last_element)
            del parser_state.token_document[-1]
        return pre_tokens

    @staticmethod
    def realize_leading_whitespace(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        original_line: str,
    ) -> str:
        """
        In cases where we "probably" have more than 3 spaces, we need to check to make
        sure that we actually have those once the containers are taken into account.
        """
        new_whitespace = extracted_whitespace
        if (
            extracted_whitespace
            and position_marker.index_indent
            # and len(extracted_whitespace) > 3
        ):
            indexed_original_line = original_line[: position_marker.index_indent]
            if indexed_original_line.endswith(">") or indexed_original_line.endswith(
                "> "
            ):
                bq_present = ParserHelper.count_characters_in_text(
                    indexed_original_line, ">"
                )
                stack_index = LeafBlockHelper.__find_nth_block_quote(
                    parser_state, bq_present
                )
                new_stack_index = stack_index + 1
                best_indent = None
                while (
                    new_stack_index < len(parser_state.token_stack)
                    and parser_state.token_stack[new_stack_index].is_list
                ):
                    inner_list_token = cast(
                        ListStackToken, parser_state.token_stack[new_stack_index]
                    )
                    indent_delta = (
                        inner_list_token.indent_level - position_marker.index_indent
                    )
                    # NOTE: this assert should be triggered by a currently disabled test
                    assert indent_delta > len(extracted_whitespace)
                    # if not indent_delta > len(extracted_whitespace):
                    #     break
                    new_stack_index += 1
                new_whitespace = (
                    extracted_whitespace[best_indent:]
                    if best_indent is not None
                    else ""
                )
        return new_whitespace

    @staticmethod
    def __find_nth_block_quote(parser_state: ParserState, bq_present: int) -> int:
        bq_encountered = 0
        stack_index = 1
        while True:
            if parser_state.token_stack[stack_index].is_block_quote:
                bq_encountered += 1
                if bq_encountered == bq_present:
                    break
            stack_index += 1
        return stack_index



"""
Module to implement a plugin that ensures that the first line in a file is a top level heading.
"""

import re
from typing import List, Optional, cast

from pymarkdown.extensions.front_matter_markdown_token import FrontMatterMarkdownToken
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.atx_heading_markdown_token import AtxHeadingMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd041(RulePlugin):
    """
    Class to implement a plugin that ensures that the first line in a file is a top level heading.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__start_level: int = -1
        self.__front_matter_title: str = ""
        self.__have_seen_first_token: bool = False
        self.__seen_html_block_start: Optional[MarkdownToken] = None
        self.__wait_for_html_end = False
        self.__html_tags_to_skip: List[str] = []

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="first-line-heading,first-line-h1",
            plugin_id="MD041",
            plugin_enabled_by_default=True,
            plugin_description="First line in file should be a top level heading",
            plugin_version="0.7.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md041.md",
            plugin_configuration="level,front_matter_title",
        )

    @classmethod
    def __validate_configuration_level(cls, found_value: int) -> None:
        if found_value < 1 or found_value > 6:
            raise ValueError("Allowable values are between 1 and 6.")

    @classmethod
    def __validate_configuration_title(cls, found_value: str) -> None:
        found_value = found_value.strip(" ")
        if ":" in found_value:
            raise ValueError("Colons (:) are not allowed in the value.")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__start_level = (
            self.plugin_configuration.get_integer_property_with_default(
                "level",
                1,
                valid_value_fn=self.__validate_configuration_level,
            )
        )
        self.__front_matter_title = (
            self.plugin_configuration.get_string_property_with_default(
                "front_matter_title",
                "title",
                valid_value_fn=self.__validate_configuration_title,
            )
            .lower()
            .strip(" ")
        )

        self.__html_tags_to_skip = ["<!--"]
        eligible_tags = self.plugin_configuration.get_string_property_with_default(
            "invisible_tags",
            "link",
            valid_value_fn=self.__validate_configuration_title,
        )

        for next_tag in eligible_tags.split(","):
            next_tag = next_tag.strip(" ")
            if next_tag == "":
                raise ValueError("Empty tag name found in between commas.")
            if re.match(r"^[a-zA-Z][a-zA-Z0-9]*$", next_tag) is None:
                raise ValueError(f"Invalid tag name '{next_tag}' found between commas.")
            self.__html_tags_to_skip.append(f"<{next_tag}")

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("level", self.__start_level),
            QueryConfigItem("front_matter_title", self.__front_matter_title),
            QueryConfigItem("invisible_tags", ",".join(self.__html_tags_to_skip)),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__have_seen_first_token = False
        self.__seen_html_block_start = None
        self.__wait_for_html_end = False
        self.__wait_for_html_end = False

    def __handle_heading_start(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        atx_token = cast(AtxHeadingMarkdownToken, token)
        self.__have_seen_first_token = True
        if atx_token.hash_count != self.__start_level:
            self.report_next_token_error(
                context, token, use_original_position=token.is_setext_heading
            )
        self.__have_seen_first_token = True

    def __handle_front_matter(self, token: MarkdownToken) -> None:
        front_token = cast(FrontMatterMarkdownToken, token)
        if self.__front_matter_title in front_token.matter_map:
            self.__have_seen_first_token = True

    def __handle_html_block_started(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        assert token.is_text
        text_token = cast(TextMarkdownToken, token)
        html_block_contents = text_token.token_text.strip(" ")

        tag_name = re.split(r"[ \n\>]", html_block_contents)[0].lower()
        if tag_name in self.__html_tags_to_skip:
            self.__wait_for_html_end = True
            return

        if not html_block_contents.startswith(
            "<h1 "
        ) and not html_block_contents.startswith("<h1>"):
            assert self.__seen_html_block_start is not None
            self.report_next_token_error(context, self.__seen_html_block_start)
        self.__have_seen_first_token = True

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if self.__have_seen_first_token:
            return
        if self.__wait_for_html_end:
            if token.is_html_block_end:
                self.__wait_for_html_end = False
                self.__seen_html_block_start = None
            return
        if token.is_atx_heading or token.is_setext_heading:
            self.__handle_heading_start(context, token)
        elif token.is_front_matter and self.__front_matter_title:
            self.__handle_front_matter(token)
        elif token.is_html_block:
            self.__seen_html_block_start = token
        elif self.__seen_html_block_start:
            self.__handle_html_block_started(context, token)
        elif not token.is_blank_line and not token.is_end_of_stream:
            self.report_next_token_error(context, token)
            self.__have_seen_first_token = True



"""
Module to provide for an encapsulation of the inline email autolink element.
"""

from typing import Optional, cast

from pymarkdown.tokens.inline_markdown_token import InlineMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class EmailAutolinkMarkdownToken(InlineMarkdownToken):
    """
    Class to provide for an encapsulation of the inline email autolink element.
    """

    def __init__(
        self,
        autolink_text: str,
        line_number: int,
        column_number: int,
        add_angle_brackets: bool = True,
    ) -> None:
        self.__autolink_text = autolink_text
        self.__add_angle_brackets = add_angle_brackets
        InlineMarkdownToken.__init__(
            self,
            MarkdownToken._token_inline_email_autolink,
            self.__autolink_text,
            line_number=line_number,
            column_number=column_number,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_inline_email_autolink

    # pylint: enable=protected-access

    @property
    def autolink_text(self) -> str:
        """
        Returns the text that is the autolink.
        """
        return self.__autolink_text

    @property
    def add_angle_brackets(self) -> bool:
        """
        Returns whether to add angle brackets.
        """
        return self.__add_angle_brackets

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            EmailAutolinkMarkdownToken,
            EmailAutolinkMarkdownToken.__rehydrate_inline_email_autolink,
            None,
        )

    @staticmethod
    def __rehydrate_inline_email_autolink(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the email autolink from the token.
        """
        _ = previous_token

        current_email_token = cast(EmailAutolinkMarkdownToken, current_token)
        prefix_char = "<"
        suffix_char = ">"
        if not current_email_token.add_angle_brackets:
            prefix_char = ""
            suffix_char = ""
        return (
            ""
            if context.block_stack[-1].is_inline_link
            else f"{prefix_char}{current_email_token.autolink_text}{suffix_char}"
        )

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            EmailAutolinkMarkdownToken,
            EmailAutolinkMarkdownToken.__handle_email_autolink_token,
            None,
        )

    @classmethod
    def __handle_email_autolink_token(
        cls,
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = transform_state

        email_token = cast(EmailAutolinkMarkdownToken, next_token)
        return "".join(
            [
                output_html,
                '<a href="mailto:',
                email_token.autolink_text,
                '">',
                email_token.autolink_text,
                "</a>",
            ]
        )



"""
Module to handle the processing of configuration for the application.
"""

import argparse
import logging
from typing import Callable, Optional

from application_properties import ApplicationProperties
from application_properties.multisource_configuration_loader import (
    ConfigurationFileType,
    MultisourceConfigurationLoader,
    MultisourceConfigurationLoaderOptions,
)

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class ApplicationConfigurationHelper:
    """
    Class to handle the processing of configuration for the application.
    """

    __pyproject_section_header = "tool.pymarkdown"
    __default_configuration_file = ".pymarkdown"

    @staticmethod
    def apply_configuration_layers(
        args: argparse.Namespace,
        properties: ApplicationProperties,
        handle_error: Callable[[str, Optional[Exception]], None],
    ) -> None:
        """
        Apply any general python configuration files followed by any configuration
        files specific to this project.
        """

        options = MultisourceConfigurationLoaderOptions(
            load_json_files_as_json5=args.use_json5_for_configuration,
            section_header_if_toml=ApplicationConfigurationHelper.__pyproject_section_header,
        )
        loader = MultisourceConfigurationLoader(options)
        loader.add_local_pyproject_toml_file(
            ApplicationConfigurationHelper.__pyproject_section_header
        )
        loader.add_local_project_configuration_file(
            ApplicationConfigurationHelper.__default_configuration_file,
            ConfigurationFileType.JSON,
            [ConfigurationFileType.YAML, ConfigurationFileType.YML],
        )
        if args.configuration_file:
            loader.add_specified_configuration_file(
                args.configuration_file, ConfigurationFileType.NONE
            )
        loader.add_manually_set_properties(args.set_configuration)
        loader.process(properties, handle_error)

        if args.strict_configuration or properties.get_boolean_property(
            "mode.strict-config", strict_mode=True
        ):
            properties.enable_strict_mode()


# pylint: enable=too-few-public-methods



"""
Module to implement a plugin that looks for leading and trailing spaces within code spans.
"""

from typing import cast

from pymarkdown.plugin_manager.plugin_details import PluginDetailsV2
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.inline_code_span_markdown_token import (
    InlineCodeSpanMarkdownToken,
)
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd038(RulePlugin):
    """
    Class to implement a plugin that looks for leading and trailing spaces within code spans.
    """

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV2(
            plugin_name="no-space-in-code",
            plugin_id="MD038",
            plugin_enabled_by_default=True,
            plugin_description="Spaces inside code span elements",
            plugin_version="0.5.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md038.md",
            plugin_supports_fix=True,
        )

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if not token.is_inline_code_span:
            return
        code_span_token = cast(InlineCodeSpanMarkdownToken, token)
        if len(code_span_token.span_text) == 1:
            has_leading = code_span_token.span_text[0] == " "
            has_trailing = False
        else:
            has_leading = (
                code_span_token.span_text[0] == " "
                and code_span_token.span_text[1] != "`"
            )
            has_trailing = (
                code_span_token.span_text[-1] == " "
                and code_span_token.span_text[-2] != "`"
            )
        if has_leading != has_trailing or has_leading:
            if context.in_fix_mode:
                adjusted_span_text = code_span_token.span_text
                if has_leading:
                    adjusted_span_text = adjusted_span_text[1:]
                if has_trailing:
                    adjusted_span_text = adjusted_span_text[:-1]
                self.register_fix_token_request(
                    context,
                    token,
                    "next_token",
                    "span_text",
                    adjusted_span_text,
                )
            else:
                self.report_next_token_error(context, token)



from dataclasses import dataclass
from typing import List, Optional

from pymarkdown.general.parser_helper import ParserHelper


@dataclass
class TableColumn:
    text: str
    leading_whitespace: str
    trailing_whitespace: str

    def __init__(self, text: str, is_last: bool = False):
        new_start_index, leading_whitespace = ParserHelper.collect_while_spaces(text, 0)
        assert new_start_index is not None
        assert leading_whitespace is not None
        self.leading_whitespace = leading_whitespace
        d = len(text) - 1 if (text[-1] == "|" and not is_last) else len(text)
        _, new_end_index = ParserHelper.collect_backwards_while_spaces(text, d)
        assert new_end_index is not None
        self.trailing_whitespace = text[new_end_index:]
        self.text = text[new_start_index:new_end_index]


@dataclass
class TableRow:
    extracted_whitespace: str
    trailing_whitespace: str
    columns: List[TableColumn]
    end_of_row: Optional[str]
    did_start_with_separator: bool


@dataclass(frozen=True)
class TableTuple:
    """
    Class to hold the tuple of information for creating a Table.
    """

    normalized_destination: Optional[str]
    xyz: List[TableRow]
    col_as: List[Optional[str]]



"""
Module to provide processing for the atx heading leaf blocks.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.container_blocks.container_grab_bag import ContainerGrabBag
from pymarkdown.container_blocks.container_helper import ContainerHelper
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.leaf_blocks.leaf_block_helper import LeafBlockHelper
from pymarkdown.tokens.atx_heading_markdown_token import AtxHeadingMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import ListStackToken, StackToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class AtxLeafBlockProcessor:
    """
    Class to provide processing for the atx heading leaf blocks.
    """

    __atx_character = "#"

    @staticmethod
    def is_atx_heading(
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        skip_whitespace_check: bool = False,
    ) -> Tuple[bool, Optional[int], Optional[int], Optional[str]]:
        """
        Determine whether or not an ATX Heading is about to start.
        """
        if (
            TabHelper.is_length_less_than_or_equal_to(extracted_whitespace, 3)
            or skip_whitespace_check
        ) and ParserHelper.is_character_at_index(
            line_to_parse,
            start_index,
            AtxLeafBlockProcessor.__atx_character,
        ):
            hash_count, new_index = ParserHelper.collect_while_character_verified(
                line_to_parse,
                start_index,
                AtxLeafBlockProcessor.__atx_character,
            )

            non_whitespace_index, _ = ParserHelper.collect_while_spaces(
                line_to_parse, new_index
            )
            extracted_whitespace_at_start = line_to_parse[
                new_index:non_whitespace_index
            ]

            if hash_count <= 6 and (
                extracted_whitespace_at_start
                or non_whitespace_index == len(line_to_parse)
            ):
                return (
                    True,
                    non_whitespace_index,
                    hash_count,
                    extracted_whitespace_at_start,
                )
        return False, None, None, None

    # pylint: disable=too-many-arguments
    @staticmethod
    def parse_atx_headings(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        block_quote_data: BlockQuoteData,
        original_line: str,
        grab_bag: ContainerGrabBag,
    ) -> List[MarkdownToken]:
        """
        Handle the parsing of an atx heading.
        """

        ex_ws = LeafBlockHelper.realize_leading_whitespace(
            parser_state, position_marker, extracted_whitespace, original_line
        )
        (
            heading_found,
            non_whitespace_index,
            hash_count,
            extracted_whitespace_at_start,
        ) = AtxLeafBlockProcessor.is_atx_heading(
            position_marker.text_to_parse,
            position_marker.index_number,
            ex_ws,
        )
        if not heading_found:
            POGGER.debug(
                "parse_atx_headings>>not eligible",
            )
            return []

        assert (
            non_whitespace_index is not None
            and hash_count is not None
            and extracted_whitespace_at_start is not None
        ), "If a heading was found, all of the rest are defined."
        return AtxLeafBlockProcessor.__parse_atx_heading_found(
            parser_state,
            position_marker,
            original_line,
            hash_count,
            non_whitespace_index,
            block_quote_data,
            extracted_whitespace_at_start,
            extracted_whitespace,
            grab_bag,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __parse_atx_heading_found(
        parser_state: ParserState,
        position_marker: PositionMarker,
        original_line: str,
        hash_count: int,
        non_whitespace_index: int,
        block_quote_data: BlockQuoteData,
        extracted_whitespace_at_start: str,
        extracted_whitespace: str,
        grab_bag: ContainerGrabBag,
    ) -> List[MarkdownToken]:
        POGGER.debug(
            "parse_atx_headings>>start",
        )

        (
            old_top_of_stack,
            remaining_line,
            remove_trailing_count,
            extracted_whitespace_before_end,
            extracted_whitespace_at_end,
            new_tokens,
            extracted_whitespace_at_start,
            extracted_whitespace,
            delay_tab_match,
        ) = AtxLeafBlockProcessor.__prepare_for_create_atx_heading(
            parser_state,
            position_marker,
            position_marker.text_to_parse[non_whitespace_index:],
            original_line,
            extracted_whitespace_at_start,
            extracted_whitespace,
            hash_count,
            block_quote_data,
            grab_bag,
        )

        POGGER.debug("extracted_whitespace>>$<<", extracted_whitespace)
        POGGER.debug("removed_chars_at_start>>$", position_marker.index_indent)
        POGGER.debug("delay_tab_match>>$", delay_tab_match)
        if delay_tab_match:
            extracted_whitespace = (
                AtxLeafBlockProcessor.__parse_atx_headings_delay_tab_match(
                    position_marker, original_line
                )
            )

        AtxLeafBlockProcessor.__parse_atx_heading_add_tokens(
            parser_state,
            position_marker,
            new_tokens,
            hash_count,
            remove_trailing_count,
            extracted_whitespace,
            old_top_of_stack,
            delay_tab_match,
            remaining_line,
            extracted_whitespace_at_start,
            extracted_whitespace_at_end,
            extracted_whitespace_before_end,
            block_quote_data,
        )
        return new_tokens

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments

    @staticmethod
    def __parse_atx_heading_add_tokens(
        parser_state: ParserState,
        position_marker: PositionMarker,
        new_tokens: List[MarkdownToken],
        hash_count: int,
        remove_trailing_count: int,
        extracted_whitespace: str,
        old_top_of_stack: StackToken,
        delay_tab_match: bool,
        remaining_line: str,
        extracted_whitespace_at_start: str,
        extracted_whitespace_at_end: str,
        extracted_whitespace_before_end: str,
        block_quote_data: BlockQuoteData,
    ) -> List[MarkdownToken]:
        start_token = AtxHeadingMarkdownToken(
            hash_count,
            remove_trailing_count,
            extracted_whitespace,
            position_marker,
        )
        new_tokens.append(start_token)

        LeafBlockHelper.correct_for_leaf_block_start_in_list(
            parser_state,
            position_marker.index_indent,
            old_top_of_stack,
            new_tokens,
            block_quote_data,
            was_token_already_added_to_stack=False,
            delay_tab_match=delay_tab_match,
        )

        new_tokens.append(
            TextMarkdownToken(
                remaining_line,
                extracted_whitespace_at_start,
                position_marker=position_marker,
            )
        )
        end_token = start_token.generate_close_markdown_token_from_markdown_token(
            extracted_whitespace_at_end, extracted_whitespace_before_end
        )
        new_tokens.append(end_token)
        return new_tokens

    # pylint: enable=too-many-arguments

    @staticmethod
    def __parse_atx_headings_delay_tab_match(
        position_marker: PositionMarker, original_line: str
    ) -> str:
        _, ex_ws = ParserHelper.extract_spaces_verified(original_line, 0)
        POGGER.debug("ex_ws>>$<<", ex_ws)

        loop_index = 0
        rep = True
        while rep and loop_index < len(ex_ws):
            loop_index += 1
            ex_ws_to_index = ex_ws[:loop_index]
            POGGER.debug("ex_ws_to_index>>$<<", ex_ws_to_index)
            ex_ws_to_index_untabbified = TabHelper.detabify_string(ex_ws_to_index)
            POGGER.debug("ex_ws_to_index_untabbified>>$<<", ex_ws_to_index_untabbified)
            POGGER.debug(
                "removed_chars=$ < len(ex_ws_to_index_untabbified)=$",
                position_marker.index_indent,
                len(ex_ws_to_index_untabbified),
            )
            rep = position_marker.index_indent > len(ex_ws_to_index_untabbified)

        POGGER.debug(
            "removed_chars=$ == len(ex_ws_to_index_untabbified)=$",
            position_marker.index_indent,
            len(ex_ws_to_index_untabbified),
        )
        extracted_whitespace = ex_ws[loop_index - 1 :]
        POGGER.debug("extracted_whitespace>>$<<", extracted_whitespace)
        return extracted_whitespace

    # pylint: disable=too-many-arguments
    @staticmethod
    def __prepare_for_create_atx_heading_with_tab(
        parser_state: ParserState,
        original_line: str,
        remaining_line: str,
        extracted_whitespace_at_start: str,
        extracted_whitespace: str,
        hash_count: int,
    ) -> Tuple[str, str, str, bool]:
        POGGER.debug(">>extracted_whitespace>:$:<", extracted_whitespace)
        POGGER.debug(
            ">>extracted_whitespace_at_start>:$:<", extracted_whitespace_at_start
        )
        reconstructed_line = (
            extracted_whitespace
            + ParserHelper.repeat_string("#", hash_count)
            + extracted_whitespace_at_start
            + remaining_line
        )

        stack_index = len(parser_state.token_stack) - 1
        while (
            stack_index != 0
            and not parser_state.token_stack[stack_index].is_list
            and not parser_state.token_stack[stack_index].is_block_quote
        ):
            stack_index -= 1
        leading_spaces = None
        if parser_state.token_stack[stack_index].is_list:
            POGGER.debug(
                ">>parser_state.token_stack[stack_index]>:$:<",
                parser_state.token_stack[stack_index],
            )
            POGGER.debug(
                ">>parser_state.token_stack[stack_index].mdt>:$:<",
                parser_state.token_stack[stack_index].matching_markdown_token,
            )
            list_markdown_token = cast(
                ListStartMarkdownToken,
                parser_state.token_stack[stack_index].matching_markdown_token,
            )
            leading_spaces = list_markdown_token.leading_spaces

            # TOoDO This needs to be fixed at a higher level, should not be needed
            # if not leading_spaces and parser_state.token_stack[stack_index].is_ordered_list:
            #     leading_spaces = list_markdown_token.extracted_whitespace
            #     assert False
            POGGER.debug(">>leading_spaces>:$:<", leading_spaces)
            if leading_spaces is not None and "\n" in leading_spaces:
                last_index = leading_spaces.rindex("\n")
                leading_spaces = leading_spaces[last_index + 1 :]

        POGGER.debug(">>reconstructed_line>:$:<", reconstructed_line)
        _, adj_original_index, split_tab = TabHelper.find_tabified_string(
            original_line,
            reconstructed_line,
            use_proper_traverse=True,
            reconstruct_prefix=leading_spaces,
        )
        POGGER.debug(">>adj_original_index>:$:<", adj_original_index)

        after_pre_hash_whitespace_index, extracted_whitespace = (
            ParserHelper.extract_spaces_verified(original_line, adj_original_index)
        )
        POGGER.debug(
            ">>after_pre_hash_whitespace_index>:$:<", after_pre_hash_whitespace_index
        )
        POGGER.debug(">>extracted_whitespace>:$:<", extracted_whitespace)

        after_post_hash_whitespace_index, extracted_whitespace_at_start = (
            ParserHelper.extract_spaces_verified(
                original_line, after_pre_hash_whitespace_index + hash_count
            )
        )
        remaining_line = original_line[after_post_hash_whitespace_index:]
        return (
            remaining_line,
            extracted_whitespace_at_start,
            extracted_whitespace,
            split_tab,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __prepare_for_create_atx_heading(
        parser_state: ParserState,
        position_marker: PositionMarker,
        remaining_line: str,
        original_line: str,
        extracted_whitespace_at_start: str,
        extracted_whitespace: str,
        hash_count: int,
        block_quote_data: BlockQuoteData,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[StackToken, str, int, str, str, List[MarkdownToken], str, str, bool]:
        (
            old_top_of_stack,
            remove_trailing_count,
        ) = (
            parser_state.token_stack[-1],
            0,
        )

        eligble_for_tab_match_delay = (
            AtxLeafBlockProcessor.__determine_eligble_for_tab_match_delay(
                parser_state, position_marker, old_top_of_stack
            )
        )

        POGGER.debug("remaining_line>:$:<", remaining_line)
        POGGER.debug("original_line>:$:<", original_line)
        if (
            ParserHelper.tab_character in original_line
            and not eligble_for_tab_match_delay
        ):
            POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
            (
                remaining_line,
                extracted_whitespace_at_start,
                extracted_whitespace,
                split_tab,
            ) = AtxLeafBlockProcessor.__prepare_for_create_atx_heading_with_tab(
                parser_state,
                original_line,
                remaining_line,
                extracted_whitespace_at_start,
                extracted_whitespace,
                hash_count,
            )
            POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
        else:
            split_tab = False
        POGGER.debug("split_tab>:$:<", split_tab)

        new_tokens, _ = parser_state.close_open_blocks_fn(parser_state)
        POGGER.debug("new_tokens>:$:<", new_tokens)
        split_tab, extracted_whitespace, _ = (
            ContainerHelper.reduce_containers_if_required(
                parser_state,
                position_marker,
                block_quote_data,
                new_tokens,
                split_tab,
                extracted_whitespace,
                grab_bag,
            )
        )
        if split_tab:
            POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
            extracted_whitespace = TabHelper.adjust_block_quote_indent_for_tab_verified(
                parser_state, extracted_whitespace
            )
            POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)

        (
            extracted_whitespace_at_end,
            extracted_whitespace_before_end,
            remaining_line,
            remove_trailing_count,
        ) = AtxLeafBlockProcessor.__prepare_for_create_atx_heading_adjust(
            remaining_line, remove_trailing_count
        )

        return (
            old_top_of_stack,
            remaining_line,
            remove_trailing_count,
            extracted_whitespace_before_end,
            extracted_whitespace_at_end,
            new_tokens,
            extracted_whitespace_at_start,
            extracted_whitespace,
            ParserHelper.tab_character in original_line and eligble_for_tab_match_delay,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __determine_eligble_for_tab_match_delay(
        parser_state: ParserState,
        position_marker: PositionMarker,
        old_top_of_stack: StackToken,
    ) -> bool:
        eligble_for_tab_match_delay = False
        keep_going = len(parser_state.token_stack) >= 2
        if keep_going:
            keep_going = False
            POGGER.debug("old_top_of_stack>:$:<", old_top_of_stack)
            POGGER.debug(
                "parser_state.token_stack[-1]>:$:<", parser_state.token_stack[-1]
            )
            POGGER.debug(
                "parser_state.token_stack[-2]>:$:<", parser_state.token_stack[-2]
            )
            if old_top_of_stack.is_paragraph:
                possible_list_index = (
                    -2 if parser_state.token_stack[-1].is_paragraph else -1
                )
                if parser_state.token_stack[possible_list_index].is_list:
                    POGGER.debug("1!!!!!")
                    keep_going = True
        if keep_going:
            POGGER.debug("2!!!!!")
            assert parser_state.token_stack[
                possible_list_index
            ].is_list, "Should follow through from above."
            list_stack_token = cast(
                ListStackToken, parser_state.token_stack[possible_list_index]
            )

            removed_chars_at_start = position_marker.index_indent
            POGGER.debug(">>removed_chars_at_start>>$>>", removed_chars_at_start)
            POGGER.debug(">>stack indent>>$>>", list_stack_token.indent_level)
            if removed_chars_at_start < list_stack_token.indent_level:
                eligble_for_tab_match_delay = True
                POGGER.debug("3!!!!!")
        return eligble_for_tab_match_delay

    @staticmethod
    def __prepare_for_create_atx_heading_adjust(
        remaining_line: str, remove_trailing_count: int
    ) -> Tuple[str, str, str, int]:
        extracted_whitespace_before_end = ""
        (
            end_index,
            extracted_whitespace_at_end,
        ) = ParserHelper.extract_spaces_from_end(remaining_line)
        while (
            end_index > 0
            and remaining_line[end_index - 1] == AtxLeafBlockProcessor.__atx_character
        ):
            end_index -= 1
            remove_trailing_count += 1
        if remove_trailing_count:
            if end_index > 0:
                if ParserHelper.is_character_at_index_whitespace(
                    remaining_line, end_index - 1
                ):
                    remaining_line = remaining_line[:end_index]
                    (
                        _,
                        end_index,
                    ) = ParserHelper.collect_backwards_while_spaces_verified(
                        remaining_line, len(remaining_line) - 1
                    )
                    extracted_whitespace_before_end = remaining_line[end_index:]
                    remaining_line = remaining_line[:end_index]
                else:
                    extracted_whitespace_at_end, remove_trailing_count = "", 0
            else:
                remaining_line = ""
        else:
            extracted_whitespace_at_end = remaining_line[end_index:]
            remaining_line = remaining_line[:end_index]
        return (
            extracted_whitespace_at_end,
            extracted_whitespace_before_end,
            remaining_line,
            remove_trailing_count,
        )



"""
Module to implement the front matter extensions.
"""

import logging
from typing import Any, List, Optional, Tuple, Union

import yaml
from yaml import SafeLoader

from pymarkdown.extension_manager.extension_impl import ExtensionDetails
from pymarkdown.extension_manager.extension_manager_constants import (
    ExtensionManagerConstants,
)
from pymarkdown.extension_manager.parser_extension import ParserExtension
from pymarkdown.extensions.front_matter_markdown_token import FrontMatterMarkdownToken
from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.source_providers import SourceProvider
from pymarkdown.leaf_blocks.thematic_leaf_block_processor import (
    ThematicLeafBlockProcessor,
)
from pymarkdown.my_application_properties_facade import MyApplicationPropertiesFacade
from pymarkdown.tokens.markdown_token import MarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class FrontMatterExtension(ParserExtension):
    """
    Extension to implement the front matter extensions.
    """

    def __init__(self) -> None:
        self.__allow_blank_lines = False

    def get_identifier(self) -> str:
        """
        Get the identifier associated with this extension.
        """
        return "front-matter"

    def get_details(self) -> ExtensionDetails:
        """
        Get the details for the extension.
        """
        return ExtensionDetails(
            extension_id=self.get_identifier(),
            extension_name="Front Matter Metadata",
            extension_description="Allows metadata to be parsed from document front matter.",
            extension_enabled_by_default=False,
            extension_version="0.5.0",
            extension_interface_version=ExtensionManagerConstants.EXTENSION_INTERFACE_VERSION_BASIC,
            extension_url="https://pymarkdown.readthedocs.io/en/latest/extensions/front-matter/",
            extension_configuration=None,
        )

    def apply_configuration(
        self, extension_specific_facade: MyApplicationPropertiesFacade
    ) -> None:
        """
        Apply any configuration required by the extension.
        """
        self.__allow_blank_lines = (
            extension_specific_facade.get_boolean_property_with_default(
                "allow_blank_lines", False
            )
        )

    # pylint: disable=too-many-arguments
    def process_header_if_present(
        self,
        first_line_in_document: str,
        line_number: int,
        requeue: List[str],
        source_provider: SourceProvider,
        tokenized_document: List[MarkdownToken],
    ) -> Tuple[Optional[str], int, List[str]]:
        """
        Take care of processing eligibility and processing for front matter support.
        """
        start_char, extracted_index = ThematicLeafBlockProcessor.is_thematic_break(
            first_line_in_document.rstrip(Constants.ascii_whitespace),
            0,
            "",
            whitespace_allowed_between_characters=False,
            skip_whitespace_check=True,
        )
        next_line: Optional[str] = first_line_in_document
        if start_char == "-" and extracted_index == 3:
            assert next_line is not None
            (
                next_line,
                new_token,
                line_number,
                requeue_lines,
            ) = self.__handle_document_front_matter(next_line, source_provider)
            if new_token:
                tokenized_document.append(new_token)
            else:
                assert requeue_lines is not None
                requeue.extend(requeue_lines)
                next_line = requeue[0]
                del requeue[0]
            POGGER.debug("self.tokenized_document>>$", tokenized_document)
            POGGER.debug("requeue>>$", requeue)
        return next_line, line_number, requeue

    # pylint: enable=too-many-arguments

    def __handle_document_front_matter(
        self, token_to_use: str, source_provider: SourceProvider
    ) -> Tuple[
        Optional[str], Optional[FrontMatterMarkdownToken], int, Optional[List[str]]
    ]:
        starting_line = token_to_use
        clean_starting_line = starting_line.rstrip(Constants.ascii_whitespace)
        repeat_again = True
        have_closing = False
        collected_lines: List[str] = []
        POGGER.info("Metadata prefix detected, scanning for metadata header.")
        next_line = None
        while repeat_again:
            next_line = source_provider.get_next_line()
            if next_line and next_line.rstrip(Constants.ascii_whitespace):
                start_char, _ = ThematicLeafBlockProcessor.is_thematic_break(
                    next_line.rstrip(Constants.ascii_whitespace),
                    0,
                    "",
                    whitespace_allowed_between_characters=False,
                )
                have_closing = bool(
                    start_char
                ) and clean_starting_line == next_line.rstrip(
                    Constants.ascii_whitespace
                )
                repeat_again = not have_closing
            elif not self.__allow_blank_lines:
                repeat_again = False
                assert next_line is not None
                collected_lines.append(next_line)
            if repeat_again:
                assert next_line is not None
                collected_lines.append(next_line)

        if not have_closing:
            POGGER.info(
                "Metadata prefix abandoned. End of document reached before closing fence encountered."
            )
            collected_lines.insert(0, starting_line)
            return None, None, 1, collected_lines

        POGGER.info("Metadata prefix collected. Verifying validity.")
        matter_map = FrontMatterExtension.__validate_yaml(collected_lines)
        POGGER.debug("ret=$s,type=$s", matter_map, type(matter_map))
        if matter_map is None or isinstance(matter_map, str):
            POGGER.info("Metadata validation failed: $", matter_map)
            collected_lines.insert(0, starting_line)
            collected_lines.append(starting_line)
            return None, None, 1, collected_lines

        POGGER.info("Metadata validation succeeded.")
        position_marker = PositionMarker(1, 0, starting_line)
        assert next_line is not None
        new_token = FrontMatterMarkdownToken(
            starting_line, next_line, collected_lines, matter_map, position_marker
        )
        return (
            source_provider.get_next_line(),
            new_token,
            3 + len(collected_lines),
            None,
        )

    @staticmethod
    def __validate_yaml(collected_lines: List[str]) -> Union[Any, str]:
        loaded_document = None
        try:
            # print(collected_lines)
            joined_lines = "\n".join(collected_lines)
            loaded_document = yaml.load(joined_lines, SafeLoader)
            did_load_as_yaml = not isinstance(loaded_document, str)
        except yaml.MarkedYAMLError:
            did_load_as_yaml = False

        # This is specifically to trigger test_front_matter_20.
        assert not (
            loaded_document
            and did_load_as_yaml
            and "test" in loaded_document
            and loaded_document["test"] == "assert"
        )

        if did_load_as_yaml:
            did_load_as_yaml = loaded_document is not None

        return (
            loaded_document
            if did_load_as_yaml
            else "Front matter was not parseable as valid YAML."
        )



"""
Module to provide for a list item that can be check off.
"""

from typing import Optional, cast

from pymarkdown.extension_manager.extension_impl import ExtensionDetails
from pymarkdown.extension_manager.extension_manager_constants import (
    ExtensionManagerConstants,
)
from pymarkdown.extension_manager.parser_extension import ParserExtension
from pymarkdown.my_application_properties_facade import MyApplicationPropertiesFacade
from pymarkdown.tokens.inline_markdown_token import InlineMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class MarkdownTaskListItemsExtension(ParserExtension):
    """
    Extension to implement the task list items extension.
    """

    def get_identifier(self) -> str:
        """
        Get the identifier associated with this extension.
        """
        return "markdown-task-list-items"

    def get_details(self) -> ExtensionDetails:
        """
        Get the details for the extension.
        """
        return ExtensionDetails(
            extension_id=self.get_identifier(),
            extension_name="Markdown Task List Items",
            extension_description="Allows parsing of Markdown task list items.",
            extension_enabled_by_default=False,
            extension_version="0.5.0",
            extension_interface_version=ExtensionManagerConstants.EXTENSION_INTERFACE_VERSION_BASIC,
            extension_url="https://pymarkdown.readthedocs.io/en/latest/extensions/task-list-items/",
            extension_configuration=None,
        )

    def apply_configuration(
        self, extension_specific_facade: MyApplicationPropertiesFacade
    ) -> None:
        """
        Apply any configuration required by the extension.
        """
        _ = extension_specific_facade  # pragma: no cover


class TaskListToken(InlineMarkdownToken):
    """
    Token that contains the pragmas for the document.
    """

    def __init__(
        self,
        checked_character: str,
        line_number: int,
        column_number: int,
    ) -> None:
        self.__checked_character = checked_character
        InlineMarkdownToken.__init__(
            self,
            InlineMarkdownToken._token_task_list,
            checked_character,
            line_number=line_number,
            column_number=column_number,
        )

    @property
    def checked_character(self) -> str:
        """
        Returns the character that is in the checkbox.
        """
        return self.__checked_character

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_task_list

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self,
        registration_function: RegisterMarkdownTransformHandlersProtocol,
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """

        registration_function(
            TaskListToken,
            TaskListToken.__rehydrate_task_list_token,
            None,
        )

    @staticmethod
    def __rehydrate_task_list_token(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the markdown characters from the token.
        """
        _ = previous_token, context

        task_list_token = cast(TaskListToken, current_token)
        return f"[{task_list_token.checked_character}]"

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            TaskListToken,
            TaskListToken.__handle_pragma_token,
            None,
        )

    @staticmethod
    def __handle_pragma_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = (transform_state, next_token)

        task_list_token = cast(TaskListToken, next_token)
        if task_list_token.checked_character == " ":
            output_html += '<input type="checkbox">'
        else:
            output_html += '<input checked="" type="checkbox">'
        return output_html



"""
Module to provide processing for the leaf blocks.
"""

import logging
from typing import List

from pymarkdown.container_blocks.container_grab_bag import ContainerGrabBag
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.html.html_helper import HtmlHelper
from pymarkdown.leaf_blocks.atx_leaf_block_processor import AtxLeafBlockProcessor
from pymarkdown.leaf_blocks.fenced_leaf_block_processor import FencedLeafBlockProcessor
from pymarkdown.leaf_blocks.leaf_block_helper import LeafBlockHelper
from pymarkdown.leaf_blocks.thematic_leaf_block_processor import (
    ThematicLeafBlockProcessor,
)
from pymarkdown.tokens.markdown_token import MarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class LeafBlockProcessor:
    """
    Class to provide processing for the leaf blocks.
    """

    # pylint: disable=too-many-arguments
    @staticmethod
    def is_paragraph_ending_leaf_block_start(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        original_line: str,
        index_indent: int,
    ) -> bool:
        """
        Determine whether we have a valid leaf block start.
        """

        is_thematic_break_start, _ = ThematicLeafBlockProcessor.is_thematic_break(
            line_to_parse,
            start_index,
            extracted_whitespace,
            skip_whitespace_check=True,
        )
        is_leaf_block_start = bool(is_thematic_break_start)
        POGGER.debug(
            "is_paragraph_ending_leaf_block_start>>is_theme_break>>$",
            is_leaf_block_start,
        )
        if not is_leaf_block_start:
            is_html_block_start, _ = HtmlHelper.is_html_block(
                line_to_parse,
                start_index,
                extracted_whitespace,
                parser_state.token_stack,
                parser_state.parse_properties,
            )
            is_leaf_block_start = bool(is_html_block_start)
            POGGER.debug(
                "is_paragraph_ending_leaf_block_start>>is_html_block>>$",
                is_leaf_block_start,
            )
        if not is_leaf_block_start:
            (
                is_leaf_block_start,
                _,
                _,
                _,
                _,
            ) = FencedLeafBlockProcessor.is_fenced_code_block(
                parser_state,
                line_to_parse,
                start_index,
                extracted_whitespace,
                original_line,
                index_indent,
                skip_whitespace_check=True,
            )
            POGGER.debug(
                "is_paragraph_ending_leaf_block_start>>is_fenced_code_block>>$",
                is_leaf_block_start,
            )
        if not is_leaf_block_start:
            is_leaf_block_start, _, _, _ = AtxLeafBlockProcessor.is_atx_heading(
                line_to_parse, start_index, extracted_whitespace
            )
            POGGER.debug(
                "is_paragraph_ending_leaf_block_start>>is_atx_heading>>$",
                is_leaf_block_start,
            )
        POGGER.debug(
            "is_paragraph_ending_leaf_block_start<<$",
            is_leaf_block_start,
        )
        return is_leaf_block_start

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def handle_html_block(
        parser_state: ParserState,
        position_marker: PositionMarker,
        outer_processed: bool,
        leaf_token_whitespace: str,
        new_tokens: List[MarkdownToken],
        grab_bag: ContainerGrabBag,
    ) -> bool:
        """
        Take care of the processing for html blocks.
        """

        POGGER.debug(">>position_marker>>ttp>>$>>", position_marker.text_to_parse)
        POGGER.debug(">>position_marker>>in>>$>>", position_marker.index_number)
        POGGER.debug(">>position_marker>>ln>>$>>", position_marker.line_number)
        did_adjust_block_quote = False
        if not outer_processed and not parser_state.token_stack[-1].is_html_block:
            POGGER.debug(">>html started?>>")
            old_top_of_stack = parser_state.token_stack[-1]
            (
                html_tokens,
                did_adjust_block_quote,
                alt_removed_chars_at_start,
                leaf_token_whitespace,
            ) = HtmlHelper.parse_html_block(
                parser_state,
                position_marker,
                leaf_token_whitespace,
                grab_bag.block_quote_data,
                grab_bag.original_line,
                grab_bag,
            )
            if html_tokens:
                POGGER.debug(">>html started>>")
                LeafBlockHelper.correct_for_leaf_block_start_in_list(
                    parser_state,
                    position_marker.index_indent,
                    old_top_of_stack,
                    html_tokens,
                    grab_bag.block_quote_data,
                    alt_removed_chars_at_start=alt_removed_chars_at_start,
                    is_html=True,
                    original_line=grab_bag.original_line,
                )
            new_tokens.extend(html_tokens)
        if parser_state.token_stack[-1].is_html_block:
            POGGER.debug(">>html continued>>")
            html_tokens = HtmlHelper.check_normal_html_block_end(
                parser_state,
                position_marker.text_to_parse,
                position_marker.index_number,
                leaf_token_whitespace,
                position_marker,
                grab_bag.original_line,
                did_adjust_block_quote,
            )
            assert html_tokens, "At least one token should have been produced."
            new_tokens.extend(html_tokens)
            outer_processed = True
        else:
            POGGER.debug(">>html not encountered>>")

        return outer_processed

    # pylint: enable=too-many-arguments

    @staticmethod
    def close_indented_block_if_indent_not_there(
        parser_state: ParserState, leaf_token_whitespace: str
    ) -> List[MarkdownToken]:
        """
        If we have an indented block going on and the current line does not
        support continuing that block, close it.
        """

        POGGER.debug(
            "__close_indented_block_if_indent_not_there>>$>",
            parser_state.token_stack[-1],
        )
        POGGER.debug("leaf_token_whitespace>>$>", leaf_token_whitespace)
        pre_tokens: List[MarkdownToken] = []
        if parser_state.token_stack[
            -1
        ].is_indented_code_block and TabHelper.is_length_less_than_or_equal_to(
            leaf_token_whitespace, 3
        ):
            pre_tokens.append(
                parser_state.token_stack[
                    -1
                ].generate_close_markdown_token_from_stack_token()
            )
            del parser_state.token_stack[-1]

            extracted_blank_line_tokens = (
                LeafBlockHelper.extract_markdown_tokens_back_to_blank_line(
                    parser_state, False
                )
            )
            extracted_blank_line_tokens.reverse()
            pre_tokens.extend(extracted_blank_line_tokens)
        POGGER.debug(
            "__close_indented_block_if_indent_not_there>>pre_tokens>$>", pre_tokens
        )
        return pre_tokens



"""
Module to provide for a leaf element that can be added to markdown parsing stream.
"""

from typing import Optional, Union

from typing_extensions import override

from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.markdown_token import MarkdownToken, MarkdownTokenClass


class LeafMarkdownToken(MarkdownToken):
    """
    Class to provide for a leaf element that can be added to markdown parsing stream.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        token_name: str,
        extra_data: Optional[str],
        line_number: int = 0,
        column_number: int = 0,
        position_marker: Optional[PositionMarker] = None,
        extracted_whitespace: str = "",
        is_extension: bool = False,
        requires_end_token: bool = False,
        can_force_close: bool = True,
    ) -> None:
        self.__extracted_whitespace = extracted_whitespace
        MarkdownToken.__init__(
            self,
            token_name,
            MarkdownTokenClass.LEAF_BLOCK,
            extra_data,
            line_number=line_number,
            column_number=column_number,
            position_marker=position_marker,
            is_extension=is_extension,
            requires_end_token=requires_end_token,
            can_force_close=can_force_close,
        )

    # pylint: enable=too-many-arguments

    @property
    def extracted_whitespace(self) -> str:
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "extracted_whitespace" and isinstance(field_value, str):
            self.__extracted_whitespace = field_value
            return True
        return False



"""
Module to contain information on link titles.
"""

from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class LinkReferenceTitles:
    """
    Class to contain information on link titles.
    """

    inline_link: Optional[str]
    inline_title: Optional[str]



"""
Module to implement a plugin that looks for multiple heading lines with the same
content.
"""

from typing import Dict, List, Optional, cast

from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.atx_heading_markdown_token import AtxHeadingMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd024(RulePlugin):
    """
    Class to implement a plugin that looks for multiple heading lines with the same
    content.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__heading_text: Optional[str] = None
        self.__start_token: Optional[MarkdownToken] = None
        self.__hash_count: int = -1
        self.__last_hash_count: int = 0
        self.__siblings_only: bool = False
        self.__heading_content_map: List[Dict[str, str]] = []

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="no-duplicate-heading,no-duplicate-header",
            plugin_id="MD024",
            plugin_enabled_by_default=True,
            plugin_description="Multiple headings cannot contain the same content.",
            plugin_version="0.6.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md024.md",
            plugin_configuration="siblings_only, allow_different_nesting",
        )

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__siblings_only = (
            self.plugin_configuration.get_boolean_property_with_default(
                "siblings_only", False
            )
            or self.plugin_configuration.get_boolean_property_with_default(
                "allow_different_nesting", False
            )
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("siblings_only", self.__siblings_only),
            QueryConfigItem("allow_different_nesting", self.__siblings_only),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__heading_text = None
        self.__start_token = None
        self.__hash_count = -1
        self.__last_hash_count = 0
        self.__heading_content_map = (
            [{}, {}, {}, {}, {}, {}] if self.__siblings_only else [{}]
        )

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        skip_this_token = False
        if token.is_setext_heading or token.is_atx_heading:
            atx_heading_token = cast(AtxHeadingMarkdownToken, token)
            self.handle_heading_start(atx_heading_token)
            skip_this_token = True
        elif token.is_setext_heading_end or token.is_atx_heading_end:
            self.handler_heading_end(context, token)

        if not skip_this_token and self.__heading_text is not None:
            new_token_debug = token.debug_string(include_column_row_info=False)
            if token.is_text and not self.__heading_text:
                text_token = cast(TextMarkdownToken, token)
                suffix_to_look_for = f":{text_token.extracted_whitespace}]"
                if new_token_debug.endswith(suffix_to_look_for):
                    new_token_debug = f"{new_token_debug[:-len(suffix_to_look_for)]}:]"
            self.__heading_text += new_token_debug

    def handle_heading_start(self, token: AtxHeadingMarkdownToken) -> None:
        """
        Process the start heading token, atx or setext
        """
        self.__heading_text = ""
        self.__start_token = token
        self.__hash_count = token.hash_count if self.__siblings_only else 1

    def handler_heading_end(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        """
        Process the end heading token, atx or setext
        """

        if self.__last_hash_count:
            while self.__last_hash_count < self.__hash_count:
                self.__last_hash_count += 1
                self.__heading_content_map[self.__last_hash_count - 1] = {}
            while self.__last_hash_count > self.__hash_count:
                self.__heading_content_map[self.__last_hash_count - 1] = {}
                self.__last_hash_count -= 1

        past_headings_map = self.__heading_content_map[self.__hash_count - 1]

        assert self.__heading_text is not None
        if self.__heading_text in past_headings_map:
            assert self.__start_token is not None
            self.report_next_token_error(
                context,
                self.__start_token,
                use_original_position=token.is_setext_heading_end,
            )
        else:
            past_headings_map[self.__heading_text] = self.__heading_text
        self.__heading_text = None
        self.__last_hash_count = self.__hash_count



"""
Module to provide for an encapsulation of the new list item element.
"""

from typing import Union

from typing_extensions import override

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.container_markdown_token import ContainerMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    RegisterHtmlTransformHandlersProtocol,
)


class NewListItemMarkdownToken(ContainerMarkdownToken):
    """
    Class to provide for an encapsulation of the new list item element..
    """

    def __init__(
        self,
        indent_level: int,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        list_start_content: str,
    ) -> None:
        self.__indent_level, self.__extracted_whitespace, self.__list_start_content = (
            indent_level,
            extracted_whitespace,
            list_start_content,
        )
        ContainerMarkdownToken.__init__(
            self,
            MarkdownToken._token_new_list_item,
            self.__compose_extra_data_field(),
            position_marker=position_marker,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_new_list_item

    # pylint: enable=protected-access

    @property
    def extracted_whitespace(self) -> str:
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace

    @property
    def indent_level(self) -> int:
        """
        Returns the indent level to apply for the list element.
        """
        return self.__indent_level

    @property
    def list_start_content(self) -> str:
        """
        Returns the content used to start this list element.
        """
        return self.__list_start_content

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "list_start_content" and isinstance(field_value, str):
            self.__list_start_content = field_value
            self.__compose_extra_data_field()
            return True
        if field_name == "extracted_whitespace" and isinstance(field_value, str):
            self.__extracted_whitespace = field_value
            self.__compose_extra_data_field()
            return True
        if field_name == "indent_level" and isinstance(field_value, int):
            self.__indent_level = field_value
            self.__compose_extra_data_field()
            return True
        return False

    def __compose_extra_data_field(self) -> str:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """

        new_field = MarkdownToken.extra_data_separator.join(
            [
                str(self.__indent_level),
                self.__extracted_whitespace,
                self.__list_start_content,
            ]
        )
        self._set_extra_data(new_field)
        return new_field

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            NewListItemMarkdownToken,
            NewListItemMarkdownToken.__handle_new_list_item_token,
            None,
        )

    @staticmethod
    def __handle_new_list_item_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token
        transform_state.add_trailing_text, transform_state.add_leading_text = (
            "</li>",
            "<li>",
        )
        token_parts = [output_html]
        if output_html and output_html[-1] == ">" and not output_html.endswith("</a>"):
            token_parts.append(ParserHelper.newline_character)
        return "".join(token_parts)
